from dataclasses import asdict
from typing import Any, Awaitable, Callable, Dict, List, Optional, Text

import structlog
from sanic import Blueprint, response
from sanic.request import Request, RequestParameters
from sanic.response import HTTPResponse
from twilio.twiml.voice_response import Gather, VoiceResponse

import rasa.shared.utils.io
import rasa.utils.io
from rasa.core.channels.channel import (
    CollectingOutputChannel,
    InputChannel,
    UserMessage,
    create_auth_requested_response_provider,
    requires_basic_auth,
)
from rasa.core.channels.voice_ready.utils import (
    CallParameters,
    validate_username_password_credentials,
)
from rasa.shared.core.events import BotUttered
from rasa.shared.exceptions import InvalidConfigException

logger = structlog.get_logger(__name__)


TWILIO_VOICE_PATH = "webhooks/twilio_voice/webhook"


def map_call_params(form: RequestParameters) -> CallParameters:
    """Map the Twilio Voice parameters to the CallParameters dataclass."""
    return CallParameters(
        call_id=form.get("CallSid"),
        user_phone=form.get("Caller"),
        bot_phone=form.get("Called"),
        direction=form.get("Direction"),
    )


class TwilioVoiceInput(InputChannel):
    """Input channel for Twilio Voice."""

    SUPPORTED_VOICES = [  # noqa: RUF012
        "man",
        "woman",
        "alice",
        "Polly.Mads",
        "Polly.Naja",
        "Polly.Lotte",
        "Polly.Reuben",
        "Polly.Nicole",
        "Polly.Russell",
        "Polly.Amy",
        "Polly.Brian",
        "Polly.Emma",
        "Polly.Amy-Neural",
        "Polly.Emma-Neural",
        "Polly.Brian-Neural",
        "Polly.Raveena",
        "Polly.Ivy",
        "Polly.Joanna",
        "Polly.Joey",
        "Polly.Justin",
        "Polly.Kendra",
        "Polly.Kimberly",
        "Polly.Matthew",
        "Polly.Salli",
        "Polly.Ivy-Neural",
        "Polly.Joanna-Neural",
        "Polly.Kendra-Neural",
        "Polly.Kimberly-Neural",
        "Polly.Sally-Neural",
        "Polly.Joey-Neural",
        "Polly.Justin-Neural",
        "Polly.Matthew-Neural",
        "Polly.Geraint",
        "Polly.Celine",
        "Polly.Mathieu",
        "Polly.Chantal",
        "Polly.Hans",
        "Polly.Marlene",
        "Polly.Vicki",
        "Polly.Dora",
        "Polly.Karl",
        "Polly.Carla",
        "Polly.Giorgio",
        "Polly.Mizuki",
        "Polly.Takumi",
        "Polly.Liv",
        "Polly.Jacek",
        "Polly.Jan",
        "Polly.Ewa",
        "Polly.Maja",
        "Polly.Ricardo",
        "Polly.Vitoria",
        "Polly.Camila-Neural",
        "Polly.Cristiano",
        "Polly.Ines",
        "Polly.Carmen",
        "Polly.Maxim",
        "Polly.Tatyana",
        "Polly.Conchita",
        "Polly.Enrique",
        "Polly.Miguel",
        "Polly.Penelope",
        "Polly.Lupe-Neural",
        "Polly.Astrid",
        "Polly.Filiz",
        "Polly.Gwyneth",
        "Polly.Aditi",
    ]

    SUPPORTED_SPEECH_MODELS = [  # noqa: RUF012
        "default",
        "numbers_and_commands",
        "phone_call",
    ]

    @classmethod
    def name(cls) -> Text:
        """Name of channel."""
        return "twilio_voice"

    @classmethod
    def from_credentials(cls, credentials: Optional[Dict[Text, Any]]) -> InputChannel:
        """Load custom configurations."""
        credentials = credentials or {}

        username = credentials.get("username")
        password = credentials.get("password")
        validate_username_password_credentials(username, password, "TwilioVoice")

        return cls(
            credentials.get(
                "reprompt_fallback_phrase",
                "I'm sorry I didn't get that could you rephrase.",
            ),
            credentials.get("assistant_voice", "woman"),
            credentials.get("speech_timeout", "5"),
            credentials.get("speech_model", "default"),
            credentials.get("enhanced", "false"),
            username=username,
            password=password,
        )

    def __init__(
        self,
        reprompt_fallback_phrase: Optional[Text],
        assistant_voice: Optional[Text],
        speech_timeout: Text = "5",
        speech_model: Text = "default",
        enhanced: Text = "false",
        username: Optional[Text] = None,
        password: Optional[Text] = None,
    ) -> None:
        """Creates a connection to Twilio voice.

        Args:
            reprompt_fallback_phrase: phrase to use if no user response.
            assistant_voice: name of the assistant voice to use.
            speech_timeout: how long to pause when user finished speaking.
            speech_model: type of transcription model to use from Twilio.
            enhanced: toggle to use Twilio's premium speech transcription model.
        """
        self.reprompt_fallback_phrase = reprompt_fallback_phrase
        self.assistant_voice = assistant_voice
        self.speech_timeout = speech_timeout
        self.speech_model = speech_model
        self.enhanced = enhanced
        self.username = username
        self.password = password

        self._validate_configuration()

    def _validate_configuration(self) -> None:
        """Checks that the user configurations are valid."""
        if self.assistant_voice not in self.SUPPORTED_VOICES:
            self._raise_invalid_voice_exception()

        validate_username_password_credentials(
            self.username, self.password, "TwilioVoice"
        )

        try:
            int(self.speech_timeout)
        except ValueError:
            if self.speech_timeout.lower() != "auto":
                self._raise_invalid_speech_timeout_exception()

        if self.speech_model not in self.SUPPORTED_SPEECH_MODELS:
            self._raise_invalid_speech_model_exception()

        if self.enhanced.lower() not in [
            "true",
            "false",
        ]:
            self._raise_invalid_enhanced_option_exception()

        if (
            self.enhanced.lower() == "true"
            and self.speech_model.lower() != "phone_call"
        ):
            self._raise_invalid_enhanced_speech_model_exception()

        if (
            self.speech_model.lower() != "numbers_and_commands"
            and self.speech_timeout.lower() == "auto"
        ):
            self._raise_invalid_speech_model_timeout_exception()

    def _raise_invalid_speech_model_timeout_exception(self) -> None:
        """Raises an error if incompatible speech_timeout and speech_model used."""
        raise InvalidConfigException(
            "If speech_timeout is 'auto' the speech_model must be "
            "'numbers_and_commands'. Please update your speech_model "
            "to be 'numbers_and_commands' if you would like to continue "
            "using the 'auto' speech_model."
        )

    def _raise_invalid_enhanced_option_exception(self) -> None:
        """Raises an error if an invalid value is passed to the enhanced parameter."""
        raise InvalidConfigException(
            f"The value {self.enhanced} is invalid for the enhanced parameter. "
            f"You must provide either `true` or `false` for this value."
        )

    def _raise_invalid_speech_model_exception(self) -> None:
        """Raises an error if an invalid speech_model is provided."""
        raise InvalidConfigException(
            f"The value {self.speech_model} for speech_model is invalid. "
            f"You must choose one of 'default', 'numbers_and_commands', "
            f"or 'phone_call'. Refer to the documentation for details "
            f"about the selections."
        )

    def _raise_invalid_speech_timeout_exception(self) -> None:
        """Raises an error if an invalid speech_timeout is provided."""
        raise InvalidConfigException(
            f"The vale {self.speech_timeout} is an invalid value for speech_timeout. "
            f"Only integers and 'auto' are valid entries."
        )

    def _raise_invalid_voice_exception(self) -> None:
        """Raises an error if an invalid voice is provided."""
        raise InvalidConfigException(
            f"The value {self.assistant_voice} is an invalid for assistant_voice. "
            f"Please refer to the documentation for a list of valid voices "
            f"you can use for your voice assistant."
        )

    def _raise_invalid_enhanced_speech_model_exception(self) -> None:
        """Raises error if enhanced is used with an incompatible speech_model."""
        raise InvalidConfigException(
            f"If you set enhanced to 'true' then speech_model must be 'phone_call'. "
            f"Current speech_model is: {self.speech_model}."
        )

    def blueprint(
        self, on_new_message: Callable[[UserMessage], Awaitable[None]]
    ) -> Blueprint:
        """Defines endpoints for Twilio voice channel."""
        twilio_voice_webhook = Blueprint("Twilio_voice_webhook", __name__)

        @twilio_voice_webhook.route("/", methods=["GET"])
        async def health(request: Request) -> HTTPResponse:
            return response.json({"status": "ok"})

        @twilio_voice_webhook.route("/webhook", methods=["POST"])
        @requires_basic_auth(
            username=self.username,
            password=self.password,
            auth_request_provider=create_auth_requested_response_provider(
                TWILIO_VOICE_PATH
            ),
        )
        async def receive(request: Request) -> HTTPResponse:
            sender_id = request.form.get("From")
            text = request.form.get("SpeechResult")
            input_channel = self.name()
            call_status = request.form.get("CallStatus")
            metadata = {}

            collector = TwilioVoiceCollectingOutputChannel()

            logger.debug(
                "twilio_voice.webhook",
                sender_id=sender_id,
                text=text,
                call_status=call_status,
            )
            # Provide an initial greeting to answer the user's call.
            if (text is None) and (call_status == "ringing"):
                text = "/session_start"
                metadata = asdict(map_call_params(request.form))

            # when call is disconnected
            if call_status == "completed":
                text = "/session_end"
                metadata = {"reason": "user disconnected"}

            # determine the response.
            if text is not None:
                logger.info("twilio_voice.webhook.text_not_none", sender_id=sender_id)
                await on_new_message(
                    UserMessage(
                        text,
                        collector,
                        sender_id,
                        input_channel=input_channel,
                        metadata=metadata,
                    )
                )

                twilio_response = self._build_twilio_voice_response(collector.messages)
            # If the user doesn't respond resend the last message.
            else:
                logger.info("twilio_voice.webhook.text_none", sender_id=sender_id)
                # Get last user utterance from tracker.
                tracker = await request.app.ctx.agent.tracker_store.retrieve(sender_id)
                last_response = None
                if tracker:
                    last_response = next(
                        (
                            e
                            for e in reversed(tracker.events)
                            if isinstance(e, BotUttered)
                        ),
                        None,
                    )

                # If no previous utterance found use the reprompt_fallback phrase.
                if last_response is None:
                    last_response_text = self.reprompt_fallback_phrase
                else:
                    last_response_text = last_response.text

                twilio_response = self._build_twilio_voice_response(
                    [{"text": last_response_text}]
                )

            logger.debug(
                "twilio_voice.webhook.twilio_response",
                twilio_response=str(twilio_response),
            )
            return response.text(str(twilio_response), content_type="text/xml")

        return twilio_voice_webhook

    def _build_twilio_voice_response(
        self, messages: List[Dict[Text, Any]]
    ) -> VoiceResponse:
        """Builds the Twilio Voice Response object."""
        logger.debug("twilio_voice.build_twilio_voice_response", messages=messages)
        voice_response = VoiceResponse()
        gather = Gather(
            input="speech",
            action=f"/webhooks/{self.name()}/webhook",
            actionOnEmptyResult=True,
            speechTimeout=self.speech_timeout,
            speechModel=self.speech_model,
            enhanced=self.enhanced,
        )

        if not messages:
            # In case bot has a greet message disabled
            # or if the bot is not configured to send an initial message
            # we need to send a voice response with speech settings
            voice_response.append(gather)
            return voice_response

        # Add pauses between messages.
        # Add a listener to the last message to listen for user response.
        for i, message in enumerate(messages):
            msg_text = message["text"]
            # Check if the message is a hangup message.
            if message.get("custom", {}).get("hangup"):
                voice_response.hangup()
                break

            if i + 1 == len(messages):
                gather.say(msg_text, voice=self.assistant_voice)
                voice_response.append(gather)
            else:
                voice_response.say(msg_text, voice=self.assistant_voice)
                voice_response.pause(length=1)

        return voice_response

    def _raise_invalid_credentials_exception(self) -> None:
        # This method is now redundant since we use the shared validation function
        # but keeping it for backward compatibility if any external code calls it
        validate_username_password_credentials(
            self.username, self.password, "TwilioVoice"
        )


class TwilioVoiceCollectingOutputChannel(CollectingOutputChannel):
    """Output channel that collects send messages in a list.

    (doesn't send them anywhere, just collects them).
    """

    @classmethod
    def name(cls) -> Text:
        """Name of the output channel."""
        return "twilio_voice"

    async def send_text_message(
        self, recipient_id: Text, text: Text, **kwargs: Any
    ) -> None:
        """Sends the text message after removing emojis."""
        text = rasa.utils.io.remove_emojis(text)
        for message_part in text.strip().split("\n\n"):
            await self._persist_message(self._message(recipient_id, text=message_part))

    async def send_text_with_buttons(
        self,
        recipient_id: str,
        text: str,
        buttons: List[Dict[str, Any]],
        **kwargs: Any,
    ) -> None:
        """Uses the concise button output format for voice channels."""
        await self.send_text_with_buttons_concise(recipient_id, text, buttons, **kwargs)

    async def send_image_url(
        self, recipient_id: Text, image: Text, **kwargs: Any
    ) -> None:
        """For voice channel do not send images."""
        rasa.shared.utils.io.raise_warning(
            "An image was removed from the voice message and "
            "only the text of message was sent. "
            "It's recommended that you define voice-friendly "
            "alternatives for all responses "
            "with a visual elements such as images and emojis "
            "that are used in your voice channel."
        )

    async def hangup(self, recipient_id: Text, **kwargs: Any) -> None:
        """
        Indicate that the conversation should be ended.

        Parent class is a collecting output channel, so we don't actually hang up
        but we add a custom message to the list of messages to be sent.
        This message will be picked up by _build_twilio_voice_response
        which will hang up the call.
        """
        await self._persist_message(
            self._message(recipient_id, custom={"hangup": True})
        )



import json
from typing import TYPE_CHECKING, Any, Dict, List, Optional, Tuple, Union

import jsonschema
import numpy as np
import structlog
from pydantic import BaseModel, ConfigDict

from rasa.core.constants import (
    UTTER_SOURCE_METADATA_KEY,
)
from rasa.e2e_test.constants import (
    KEY_JUSTIFICATION,
    KEY_SCORE,
)
from rasa.e2e_test.e2e_config import LLMJudgeConfig
from rasa.shared.constants import MODEL_CONFIG_KEY, OPENAI_PROVIDER, PROVIDER_CONFIG_KEY
from rasa.shared.core.events import BotUttered
from rasa.shared.exceptions import RasaException
from rasa.shared.utils.llm import DEFAULT_OPENAI_EMBEDDING_MODEL_NAME, embedder_factory

if TYPE_CHECKING:
    from rasa.shared.core.events import Event


structlogger = structlog.get_logger()


DEFAULT_EMBEDDINGS_CONFIG = {
    PROVIDER_CONFIG_KEY: OPENAI_PROVIDER,
    MODEL_CONFIG_KEY: DEFAULT_OPENAI_EMBEDDING_MODEL_NAME,
}

ELIGIBLE_UTTER_SOURCE_METADATA = [
    "EnterpriseSearchPolicy",
    "ContextualResponseRephraser",
    "IntentlessPolicy",
]

GROUNDEDNESS_JSON_SUB_SCHEMA = {
    "properties": {
        "statements": {
            "type": "array",
            "items": {
                "type": "object",
                "properties": {
                    "statement": {"type": "string"},
                    "score": {"type": "number"},
                    "justification": {"type": "string"},
                },
                "required": ["statement", "score", "justification"],
            },
        }
    },
    "required": ["statements"],
}

ANSWER_RELEVANCE_JSON_SUB_SCHEMA = {
    "properties": {
        "question_variations": {
            "type": "array",
            "items": {"type": "string"},
        },
    },
    "required": ["question_variations"],
}

LLM_JUDGE_OUTPUT_JSON_SCHEMA = {
    "type": "object",
    "oneOf": [
        GROUNDEDNESS_JSON_SUB_SCHEMA,
        ANSWER_RELEVANCE_JSON_SUB_SCHEMA,
    ],
}


class ScoreInputs(BaseModel):
    """Input data for the score calculation."""

    model_config = ConfigDict(arbitrary_types_allowed=True)

    threshold: float
    matching_event: BotUttered
    user_question: str
    llm_judge_config: LLMJudgeConfig


def _calculate_similarity(
    user_question: str,
    generated_questions: List[str],
    llm_judge_config: LLMJudgeConfig,
) -> np.ndarray:
    """Calculate the cosine similarity between the user question and the generated questions."""  # noqa: E501
    embedding_client = embedder_factory(
        llm_judge_config.embeddings_config_as_dict, DEFAULT_EMBEDDINGS_CONFIG
    )

    user_question_embedding_response = embedding_client.embed([user_question])
    question_vector = np.asarray(user_question_embedding_response.data[0]).reshape(
        1, -1
    )

    gen_questions_embedding_response = embedding_client.embed(generated_questions)
    generated_questions_vectors = np.asarray(
        gen_questions_embedding_response.data
    ).reshape(len(generated_questions), -1)

    # calculate norm
    question_vector_norm = np.linalg.norm(question_vector, axis=1)
    generated_questions_vectors_norm = np.linalg.norm(
        generated_questions_vectors, axis=1
    )
    norm = generated_questions_vectors_norm * question_vector_norm
    norm = np.maximum(norm, 1e-10)

    # calculate the dot product
    dot_product = np.dot(generated_questions_vectors, question_vector.T).reshape(-1)

    # calculate and return cosine similarity
    return dot_product / norm


def calculate_relevance_score(
    processed_output: List[Union[str, Dict[str, Any]]],
    score_inputs: ScoreInputs,
) -> Tuple[float, str]:
    """Calculate the score based on the LLM response."""
    user_question = score_inputs.user_question
    llm_judge_config = score_inputs.llm_judge_config

    generated_questions = [output for output in processed_output]
    if all(not question for question in generated_questions):
        score = 0.0
        error_justification = "No relevant questions were generated"
        return score, error_justification

    cosine_sim = _calculate_similarity(
        user_question, generated_questions, llm_judge_config
    )

    score = cosine_sim.mean()

    if score < score_inputs.threshold:
        error_justifications = [
            f"Question '{generated_questions[i]}' "
            f"has a cosine similarity score of '{round(cosine_sim[i], 2)}' "
            f"with the user question '{user_question}'"
            for i in range(len(generated_questions))
        ]
        error_justification = ", ".join(error_justifications)
        return score, error_justification

    return score, ""


def calculate_groundedness_score(
    processed_output: List[Any],
    score_inputs: ScoreInputs,
) -> Tuple[float, str]:
    """Calculate the score based on the LLM response."""
    matching_event = score_inputs.matching_event

    total_statements = len(processed_output)
    correct_statements = sum([output.get(KEY_SCORE, 0) for output in processed_output])
    score = correct_statements / total_statements

    structlogger.debug(
        "generative_response_is_grounded_assertion.run_results",
        matching_event=repr(matching_event),
        score=score,
        justification=f"There were {correct_statements} correct statements "
        f"out of {total_statements} total extracted statements.",
    )

    if score < score_inputs.threshold:
        justifications = [
            output.get(KEY_JUSTIFICATION, "")
            for output in processed_output
            if output.get(KEY_SCORE, 0) == 0
        ]
        justification = ", ".join(justifications).replace(".", "")

        error_justification = (
            f"There were {total_statements - correct_statements} "
            f"incorrect statements out of {total_statements} total "
            f"extracted statements. The justifications for "
            f"these statements include: {justification}"
        )

        return score, error_justification

    return score, ""


def _find_matching_generative_events(
    turn_events: List["Event"], utter_source: Optional[str]
) -> List[BotUttered]:
    """Find the matching events for the generative response assertions."""
    if utter_source is None:
        return [
            event
            for event in turn_events
            if isinstance(event, BotUttered)
            and event.metadata.get(UTTER_SOURCE_METADATA_KEY)
            in ELIGIBLE_UTTER_SOURCE_METADATA
        ]

    return [
        event
        for event in turn_events
        if isinstance(event, BotUttered)
        and event.metadata.get(UTTER_SOURCE_METADATA_KEY) == utter_source
    ]


def _parse_llm_output(llm_response: str, bot_message: str) -> Dict[str, Any]:
    """Parse the LLM output."""
    llm_output = (
        llm_response.replace("```json\n", "").replace("```", "").replace("\n", "")
    )
    try:
        parsed_llm_output = json.loads(llm_output)
    except json.JSONDecodeError as exc:
        raise RasaException(
            f"Failed to parse the LLM Judge response '{llm_output}' for "
            f"the generative bot message '{bot_message}': {exc}"
        )

    return parsed_llm_output


def _validate_parsed_llm_output(
    parsed_llm_output: Dict[str, Any], bot_message: str
) -> None:
    """Validate the parsed LLM output."""
    try:
        jsonschema.validate(parsed_llm_output, LLM_JUDGE_OUTPUT_JSON_SCHEMA)
    except jsonschema.ValidationError as exc:
        raise RasaException(
            f"Failed to validate the LLM Judge json response "
            f"'{parsed_llm_output}' for the generative bot message "
            f"'{bot_message}'. Error: {exc}"
        )



from __future__ import annotations

import importlib
from dataclasses import dataclass
from typing import TYPE_CHECKING, Any, List, Optional, Text

import structlog

from rasa.shared.exceptions import RasaException
from rasa.utils.endpoints import EndpointConfig

if TYPE_CHECKING:
    from langchain.schema import Document
    from langchain.schema.embeddings import Embeddings


logger = structlog.get_logger()


@dataclass
class SearchResult:
    """A search result object.

    Attributes:
        text: The text content of the retrieved document result.
        metadata: The metadata associated with the document result.
        score: The score of the search result.
    """

    text: str
    metadata: dict
    score: Optional[float] = None

    @classmethod
    def from_document(cls, document: Document) -> "SearchResult":
        """Construct a SearchResult object from Langchain Document object."""
        return cls(text=document.page_content, metadata=document.metadata)


@dataclass
class SearchResultList:
    results: List[SearchResult]
    metadata: dict

    @classmethod
    def from_document_list(cls, documents: List["Document"]) -> "SearchResultList":
        """
        Convert a list of Langchain Documents to a SearchResultList object.

        Args:
            documents: List of Langchain Documents.

        Returns:
            SearchResultList object.
        """
        return cls(
            results=[SearchResult.from_document(doc) for doc in documents],
            metadata={"total_results": len(documents)},
        )


class InformationRetrievalException(RasaException):
    """Base class for exceptions raised by InformationRetrieval operations."""

    def __init__(self) -> None:
        self.base_message = "An error occurred while searching for documents: "

    def __str__(self) -> str:
        return self.base_message + f"{self.__cause__}"


class InformationRetrieval:
    """Base class for any InformationRetrieval implementation."""

    def __init__(self, embeddings: "Embeddings") -> None:
        self.embeddings = embeddings

    def connect(
        self,
        config: EndpointConfig,
    ) -> None:
        """Connect to the InformationRetrieval system."""
        raise NotImplementedError(
            "InformationRetrieval must implement the `connect` method."
        )

    async def search(
        self,
        query: Text,
        tracker_state: dict[str, Any],
        threshold: float = 0.0,
    ) -> SearchResultList:
        """Search for a document in the InformationRetrieval system."""
        raise NotImplementedError(
            "InformationRetrieval must implement the `search` method."
        )


def create_from_endpoint_config(
    config_type: Text,
    embeddings: "Embeddings",
) -> InformationRetrieval:
    """Instantiate a vector store based on its configuration."""
    logger.debug(
        "information_retrieval.create_from_endpoint_config", config_type=config_type
    )
    if config_type == "milvus":
        from rasa.core.information_retrieval.milvus import Milvus_Store

        return Milvus_Store(embeddings=embeddings)
    elif config_type == "qdrant":
        from rasa.core.information_retrieval.qdrant import Qdrant_Store

        return Qdrant_Store(embeddings=embeddings)
    else:
        # Import the module dynamically
        try:
            module_name, class_name = config_type.rsplit(".", 1)
            module = importlib.import_module(module_name)
        except ValueError:
            logger.error(
                "information_retrieval.create_from_endpoint_config.invalid_config",
                config_type=config_type,
            )
            raise ValueError(
                f"Invalid configuration for vector store: '{config_type}'. "
                f"Expected a module path and a class name separated by a dot."
            )
        except ModuleNotFoundError:
            logger.error(
                "information_retrieval.create_from_endpoint_config.unknown_type",
                config_type=config_type,
            )
            raise ImportError(f"Cannot retrieve class from path {config_type}.")

        external_class = getattr(module, class_name)
        return external_class(embeddings=embeddings)



from typing import List

import structlog

from rasa.dialogue_understanding.commands import (
    Command,
)

structlogger = structlog.get_logger()


class CommandParserValidatorSingleton:
    """Singleton class to validate the command parser.

    This class is used to validate the command parser. It keeps track of the number of
    consecutive turns where no commands are parsed by the command parser. If the
    number of consecutive turns exceeds a certain threshold, a warning is logged.
    The prompt can use a DSL syntax that can be incompatible with the command syntax
    used by the command parser. This class helps to detect such incompatibilities.
    """

    MAX_CONSECUTIVE_TURNS_NO_COMMAND_PREDICTED = 5
    _NO_COMMAND_PREDICTED_TURN_COUNTER = 0
    _command_parser_validated = False

    @classmethod
    def get_no_command_predicted_turn_counter(cls) -> int:
        return cls._NO_COMMAND_PREDICTED_TURN_COUNTER

    @classmethod
    def should_validate_command_parser(cls) -> bool:
        return not cls._command_parser_validated

    @classmethod
    def reset_command_parser_validation(cls) -> None:
        cls._NO_COMMAND_PREDICTED_TURN_COUNTER = 0
        cls._command_parser_validated = False

    @classmethod
    def validate_if_commands_are_parsed_from_llm_response(
        cls, commands: List[Command], llm_response: str
    ) -> None:
        if llm_response and not commands:
            cls._NO_COMMAND_PREDICTED_TURN_COUNTER += 1
        else:
            # Reset the counter if commands are generated, and mark
            # the command parser as validated.
            cls._NO_COMMAND_PREDICTED_TURN_COUNTER = 0
            cls._command_parser_validated = True
            return

        if (
            cls._NO_COMMAND_PREDICTED_TURN_COUNTER
            >= cls.MAX_CONSECUTIVE_TURNS_NO_COMMAND_PREDICTED
        ):
            structlogger.warning(
                "llm_command_generator.predict_commands.command_parser_not_working",
                event_info=(
                    f"No commands were generated by the command parser for the last "
                    f"{cls._NO_COMMAND_PREDICTED_TURN_COUNTER} times. Check if you "
                    "are running incompatible prompt and LLM command generator."
                ),
            )



from __future__ import annotations

import argparse
from pathlib import Path
from typing import Dict, Text

import structlog

import rasa.shared.utils.cli
from rasa.shared.core.flows.yaml_flows_io import YamlFlowsWriter
from rasa.shared.importers.importer import TrainingDataImporter
from rasa.shared.nlu.training_data.formats.rasa_yaml import RasaYAMLWriter
from rasa.shared.utils.io import read_file
from rasa.shared.utils.yaml import dump_obj_as_yaml_to_string
from rasa.studio.config import StudioConfig
from rasa.studio.link import get_studio_config, read_assistant_name
from rasa.studio.upload import (
    build_import_request,
    make_request,
    run_validation,
)
from rasa.studio.utils import validate_argument_paths

structlogger = structlog.get_logger(__name__)


def _send_to_studio(
    assistant_name: Text,
    payload_parts: Dict[Text, Text],
    studio_cfg: StudioConfig,
) -> None:
    """Build the GraphQL request and send it.

    Args:
        assistant_name: The name of the assistant.
        payload_parts: The parts of the payload to send.
        studio_cfg: The StudioConfig object.
    """
    graphql_req = build_import_request(
        assistant_name=assistant_name,
        flows_yaml=payload_parts.get("flows"),
        domain_yaml=payload_parts.get("domain"),
        config_yaml=payload_parts.get("config"),
        endpoints=payload_parts.get("endpoints"),
        nlu_yaml=payload_parts.get("nlu"),
    )
    verify = not studio_cfg.disable_verify
    result = make_request(studio_cfg.studio_url, graphql_req, verify)
    if not result.was_successful:
        rasa.shared.utils.cli.print_error_and_exit(result.message)

    structlogger.info(
        "studio.push.success",
        event_info=f"Pushed data to assistant '{assistant_name}'.",
        assistant_name=assistant_name,
    )
    rasa.shared.utils.cli.print_success(f"Pushed data to assistant '{assistant_name}'.")


def handle_push(args: argparse.Namespace) -> None:
    """Push the entire assistant.

    Args:
        args: The command line arguments.
    """
    validate_argument_paths(args)
    studio_cfg = get_studio_config()

    run_validation(args)

    importer = TrainingDataImporter.load_from_dict(
        domain_path=args.domain,
        config_path=args.config,
        expand_env_vars=False,
    )

    domain_yaml = dump_obj_as_yaml_to_string(importer.get_user_domain().as_dict())
    config_yaml = read_file(Path(args.config))
    endpoints_yaml = read_file(Path(args.endpoints))

    flow_importer = TrainingDataImporter.load_from_dict(
        training_data_paths=args.data, expand_env_vars=False
    )
    flows_yaml = YamlFlowsWriter().dumps(flow_importer.get_user_flows())

    nlu_importer = TrainingDataImporter.load_from_dict(
        training_data_paths=args.data, expand_env_vars=False
    )
    nlu_yaml = RasaYAMLWriter().dumps(nlu_importer.get_nlu_data())

    assistant_name = read_assistant_name()
    _send_to_studio(
        assistant_name,
        {
            "flows": flows_yaml,
            "domain": domain_yaml,
            "config": config_yaml,
            "endpoints": endpoints_yaml,
            "nlu": nlu_yaml,
        },
        studio_cfg,
    )


def handle_push_config(args: argparse.Namespace) -> None:
    """Push only the assistant configuration (config.yml).

    Args:
        args: The command line arguments.
    """
    studio_cfg = get_studio_config()
    assistant_name = read_assistant_name()

    config_yaml = read_file(Path(args.config))
    if not config_yaml:
        rasa.shared.utils.cli.print_error_and_exit(
            "No configuration data was found in the assistant."
        )

    _send_to_studio(assistant_name, {"config": config_yaml}, studio_cfg)


def handle_push_endpoints(args: argparse.Namespace) -> None:
    """Push only the endpoints configuration (endpoints.yml).

    Args:
        args: The command line arguments.
    """
    studio_cfg = get_studio_config()
    assistant_name = read_assistant_name()

    endpoints_yaml = read_file(Path(args.endpoints))
    if not endpoints_yaml:
        rasa.shared.utils.cli.print_error_and_exit(
            "No endpoints data was found in the assistant."
        )

    _send_to_studio(assistant_name, {"endpoints": endpoints_yaml}, studio_cfg)



from typing import Any, Dict, List, Optional

from rasa.core.actions.action import Action
from rasa.core.channels import OutputChannel
from rasa.core.nlg import NaturalLanguageGenerator
from rasa.shared.core.constants import ACTION_HANGUP
from rasa.shared.core.domain import Domain
from rasa.shared.core.events import Event, SessionEnded
from rasa.shared.core.trackers import DialogueStateTracker


class ActionHangup(Action):
    """Action which hangs up the call."""

    def name(self) -> str:
        """Return the name of the action."""
        return ACTION_HANGUP

    async def run(
        self,
        output_channel: OutputChannel,
        nlg: NaturalLanguageGenerator,
        tracker: DialogueStateTracker,
        domain: Domain,
        metadata: Optional[Dict[str, Any]] = None,
    ) -> List[Event]:
        """Ask the output channel to hang up the call."""
        await output_channel.hangup(tracker.sender_id)
        return [SessionEnded(metadata={"_reason": "hangup"})]



from __future__ import annotations

import os
import tempfile
from typing import Any, Dict, List, Optional, Text

from pydantic import BaseModel, ConfigDict, Field

from rasa.dialogue_understanding.patterns.domain_for_patterns import (
    ContextField,
    generate_domain_for_default_patterns,
)
from rasa.shared.core.domain import Domain
from rasa.shared.importers.importer import FlowSyncImporter, TrainingDataImporter
from rasa.shared.utils.llm import SystemPrompts, get_system_default_prompts
from rasa.shared.utils.yaml import read_yaml, write_yaml


class Response(BaseModel):
    text: str = Field(..., description="The human-readable response string.")
    metadata: Optional[Dict[str, Any]] = Field(
        None, description="Optional metadata used by Rasa at runtime."
    )


class Slot(BaseModel):
    name: Text = Field(..., description="Name of the slot.")
    type: Text = Field(..., description="Type of the slot.")
    mappings: Optional[List[Dict[str, Any]]] = Field(
        None, description="Mappings for the slot."
    )
    model_config = ConfigDict(extra="allow")


class PatternFlow(BaseModel):
    description: str = Field(..., description="Description of the flow.")
    name: Text = Field(..., description="Name of the flow.")
    steps: List[Any] = Field(..., description="List of steps in the flow.")
    nlu_trigger: Optional[Any] = Field(None, description="NLU trigger for the flow.")
    persisted_slots: Optional[List[Text]] = Field(None, description="Slots to persist.")


class PatternDefaults(BaseModel):
    """Content of `default_flows_for_patterns.yml` after it is loaded from disk."""

    responses: Dict[Text, List[Response]] = Field(
        default_factory=dict,
        description="Response templates available out-of-the-box.",
    )
    slots: Dict[Text, Slot] = Field(
        default_factory=dict, description="Slot definitions used by the default flows."
    )
    flows: Dict[Text, PatternFlow] = Field(
        default_factory=dict, description="Default flows (conversation patterns)."
    )


class RasaDefaults(PatternDefaults):
    """Full set of defaults that Rasa injects when a project is created."""

    prompts: SystemPrompts = Field(..., description="Built-in system prompts.")
    actions: List[Text] = Field(..., description="Default actions shipped by Rasa.")
    intents: List[Text] = Field(..., description="Default intents shipped by Rasa.")
    contexts: Dict[Text, ContextField] = Field(
        ..., description="Context fields that can appear on the conversation stack."
    )

    class Config:
        extra = "forbid"
        validate_assignment = True


def _get_domain_from_importer(config: Dict[Text, Any]) -> Domain:
    """Get the domain from the TrainingDataImporter.

    Args:
        config: The config.yml file data.

    Returns:
        A Domain object .
    """
    # We create the file with delete=False so that it can be re-opened on
    # Windows. If delete=True, the file is opened with the _O_TEMPORARY flag
    # which blocks any second open() call.
    tmp = tempfile.NamedTemporaryFile("w+", suffix=".yml", delete=False)

    try:
        path = tmp.name

        # write_yaml() re-opens the same path. On Windows an already-open
        # handle keeps the file locked for further opens, so we close
        # the first handle before we call write_yaml().
        tmp.close()
        write_yaml(config, path)

        importer = TrainingDataImporter.load_from_config(
            domain_path=FlowSyncImporter.default_pattern_path(),
            config_path=path,
        )
        return importer.get_domain()

    finally:
        # Because we passed delete=False above, Python will not clean the file
        # To avoid leaving garbage in the temp directory after the tests run,
        # we remove it explicitly.
        try:
            os.remove(path)
        except FileNotFoundError:
            pass


def get_pattern_defaults(config: Dict[Text, Any]) -> PatternDefaults:
    """Get the default flows, responses and slots for patterns.

    Args:
        config: The config.yml file data.

    Returns:
        A PatternDefaults object containing the defaults.
    """
    domain = _get_domain_from_importer(config)
    domain_dict = domain.as_dict()

    # Make sure all the slots are exported, including the builtin slots
    domain_dict["slots"] = {slot.name: slot.to_dict() for slot in domain.slots}

    return PatternDefaults(**domain_dict)


def get_rasa_defaults(config_yaml: Text, endpoints_yaml: Text) -> RasaDefaults:
    """Get the default values for a Rasa project.

    Args:
        config_yaml: The content of the config.yml file.
        endpoints_yaml: The content of the endpoints.yml file.

    Returns:
        A RasaDefaults object containing the default values for the project.
    """
    config = read_yaml(config_yaml)
    endpoints = read_yaml(endpoints_yaml)

    prompts = get_system_default_prompts(config, endpoints)
    pattern_domain = generate_domain_for_default_patterns()
    pattern_defaults = get_pattern_defaults(config)

    return RasaDefaults(
        prompts=prompts,
        actions=pattern_domain.actions,
        intents=pattern_domain.intents,
        contexts=pattern_domain.contexts,
        responses=pattern_defaults.responses,
        slots=pattern_defaults.slots,
        flows=pattern_defaults.flows,
    )



from typing import Optional

import structlog

from rasa.shared.utils.io import raise_deprecation_warning

structlogger = structlog.get_logger()


def resolve_aliases(config: dict, deprecated_alias_mapping: dict) -> dict:
    """
    Resolve aliases in the configuration to standard keys.

    Args:
        config: Dictionary containing the configuration.
        deprecated_alias_mapping: Dictionary mapping aliases to
            their standard keys.

    Returns:
        New dictionary containing the processed configuration.
    """
    config = config.copy()

    for alias, standard_key in deprecated_alias_mapping.items():
        # We check for the alias instead of the standard key because our goal is to
        # update the standard key when the alias is found. Since the standard key is
        # always included in the default component configurations, we overwrite it
        # with the alias value if the alias exists.
        if alias in config:
            config[standard_key] = config.pop(alias)

    return config


def raise_deprecation_warnings(
    config: dict,
    deprecated_alias_mapping: dict,
    source: Optional[str] = None,
) -> None:
    """
    Raises warnings for deprecated keys in the configuration.

    Args:
        config: Dictionary containing the configuration.
        deprecated_alias_mapping: Dictionary mapping deprecated keys to
            their standard keys.

    Raises:
        DeprecationWarning: If any deprecated key is found in the config.
    """
    for alias, standard_key in deprecated_alias_mapping.items():
        if alias in config:
            source = f"{source}: " or ""
            raise_deprecation_warning(
                message=(
                    f"{source}"
                    f"'{alias}' is deprecated and will be removed in "
                    f"4.0.0. Use '{standard_key}' instead."
                )
            )


def validate_required_keys(config: dict, required_keys: list) -> None:
    """
    Validates that the passed config contains all the required keys.

    Args:
        config: Dictionary containing the configuration.
        required_keys: List of keys that must be present in the config.

    Raises:
        ValueError: If any required key is missing.
    """
    missing_keys = [key for key in required_keys if key not in config]
    if missing_keys:
        message = f"Missing required keys '{missing_keys}' for configuration."
        structlogger.error(
            "validate_required_keys",
            message=message,
            missing_keys=missing_keys,
            config=config,
        )
        raise ValueError(message)


def validate_forbidden_keys(config: dict, forbidden_keys: list) -> None:
    """
    Validates that the passed config doesn't contain any forbidden keys.

    Args:
        config: Dictionary containing the configuration.
        forbidden_keys: List of keys that are forbidden in the config.

    Raises:
        ValueError: If any forbidden key is present.
    """
    forbidden_keys_in_config = set(config.keys()).intersection(set(forbidden_keys))

    if forbidden_keys_in_config:
        message = (
            f"Forbidden keys '{forbidden_keys_in_config}' present "
            f"in the configuration."
        )
        structlogger.error(
            "validate_forbidden_keys",
            message=message,
            forbidden_keys=forbidden_keys_in_config,
            config=config,
        )
        raise ValueError(message)



from typing import Any, Dict, Optional, Text

import structlog
from jinja2 import Template

from rasa import telemetry
from rasa.core.nlg.response import TemplatedNaturalLanguageGenerator
from rasa.core.nlg.summarize import (
    _count_multiple_utterances_as_single_turn,
    summarize_conversation,
)
from rasa.shared.constants import (
    LLM_CONFIG_KEY,
    MAX_COMPLETION_TOKENS_CONFIG_KEY,
    MODEL_CONFIG_KEY,
    MODEL_GROUP_ID_CONFIG_KEY,
    MODEL_NAME_CONFIG_KEY,
    OPENAI_PROVIDER,
    PROMPT_CONFIG_KEY,
    PROMPT_TEMPLATE_CONFIG_KEY,
    PROVIDER_CONFIG_KEY,
    TEMPERATURE_CONFIG_KEY,
    TIMEOUT_CONFIG_KEY,
)
from rasa.shared.core.domain import KEY_RESPONSES_TEXT, Domain
from rasa.shared.core.events import BotUttered, UserUttered
from rasa.shared.core.trackers import DialogueStateTracker
from rasa.shared.nlu.constants import (
    KEY_COMPONENT_NAME,
    KEY_LLM_RESPONSE_METADATA,
    KEY_PROMPT_NAME,
    KEY_USER_PROMPT,
    PROMPTS,
)
from rasa.shared.providers.llm.llm_response import LLMResponse, measure_llm_latency
from rasa.shared.utils.constants import LOG_COMPONENT_SOURCE_METHOD_INIT
from rasa.shared.utils.health_check.llm_health_check_mixin import LLMHealthCheckMixin
from rasa.shared.utils.llm import (
    DEFAULT_OPENAI_GENERATE_MODEL_NAME,
    DEFAULT_OPENAI_MAX_GENERATED_TOKENS,
    USER,
    check_prompt_config_keys_and_warn_if_deprecated,
    combine_custom_and_default_config,
    get_prompt_template,
    llm_factory,
    resolve_model_client_config,
    tracker_as_readable_transcript,
)
from rasa.utils.endpoints import EndpointConfig
from rasa.utils.log_utils import log_llm

structlogger = structlog.get_logger()

RESPONSE_REPHRASING_KEY = "rephrase"

RESPONSE_REPHRASING_TEMPLATE_KEY = "rephrase_prompt"

RESPONSE_SUMMARISE_CONVERSATION_KEY = "summarize_conversation"

DEFAULT_REPHRASE_ALL = False
DEFAULT_SUMMARIZE_HISTORY = True
DEFAULT_MAX_HISTORICAL_TURNS = 5
DEFAULT_COUNT_MULTIPLE_UTTERANCES_AS_SINGLE_TURN = True

DEFAULT_LLM_CONFIG = {
    PROVIDER_CONFIG_KEY: OPENAI_PROVIDER,
    MODEL_CONFIG_KEY: DEFAULT_OPENAI_GENERATE_MODEL_NAME,
    TEMPERATURE_CONFIG_KEY: 0.3,
    MAX_COMPLETION_TOKENS_CONFIG_KEY: DEFAULT_OPENAI_MAX_GENERATED_TOKENS,
    TIMEOUT_CONFIG_KEY: 5,
}

DEFAULT_RESPONSE_VARIATION_PROMPT_TEMPLATE = """The following is a conversation with
an AI assistant. The assistant is helpful, creative, clever, and very friendly.
Rephrase the suggested AI response staying close to the original message and retaining
its meaning. Use simple {{language}}.

Context / previous conversation with the user:
{{history}}

Last user message:
{{current_input}}

Suggested AI Response: {{suggested_response}}

Rephrased AI Response:"""


class ContextualResponseRephraser(
    LLMHealthCheckMixin, TemplatedNaturalLanguageGenerator
):
    """Generates responses based on modified templates.

    The templates are filled with the entities and slots that are available in the
    tracker. The resulting response is then passed through the LLM to generate a
    variation of the response.

    The variation is only generated if the response is a text response and the
    response explicitly set `rephrase` to `True`. This is to avoid
    generating responses for templates that are to volatile to be
    modified by the LLM.

    Args:
        endpoint_config: The endpoint configuration for the LLM.
        domain: The domain of the assistant.

    Attributes:
        nlg_endpoint: The endpoint configuration for the LLM.
    """

    def __init__(self, endpoint_config: EndpointConfig, domain: Domain) -> None:
        super().__init__(domain.responses)

        self.nlg_endpoint = endpoint_config

        # Warn if the prompt config key is used to set the prompt template
        check_prompt_config_keys_and_warn_if_deprecated(
            self.nlg_endpoint.kwargs, "contextual_response_rephraser"
        )

        self.prompt_template = get_prompt_template(
            self.nlg_endpoint.kwargs.get(PROMPT_TEMPLATE_CONFIG_KEY)
            or self.nlg_endpoint.kwargs.get(PROMPT_CONFIG_KEY),
            DEFAULT_RESPONSE_VARIATION_PROMPT_TEMPLATE,
            log_source_component=ContextualResponseRephraser.__name__,
            log_source_method=LOG_COMPONENT_SOURCE_METHOD_INIT,
        )
        self.rephrase_all = self.nlg_endpoint.kwargs.get(
            "rephrase_all", DEFAULT_REPHRASE_ALL
        )
        self.trace_prompt_tokens = self.nlg_endpoint.kwargs.get(
            "trace_prompt_tokens", False
        )
        self.summarize_history = self.nlg_endpoint.kwargs.get(
            "summarize_history", DEFAULT_SUMMARIZE_HISTORY
        )
        self.max_historical_turns = self.nlg_endpoint.kwargs.get(
            "max_historical_turns", DEFAULT_MAX_HISTORICAL_TURNS
        )

        self.count_multiple_utterances_as_single_turn = self.nlg_endpoint.kwargs.get(
            "count_multiple_utterances_as_single_turn",
            DEFAULT_COUNT_MULTIPLE_UTTERANCES_AS_SINGLE_TURN,
        )

        self.llm_config = resolve_model_client_config(
            self.nlg_endpoint.kwargs.get(LLM_CONFIG_KEY),
            ContextualResponseRephraser.__name__,
        )

        self.perform_llm_health_check(
            self.llm_config,
            DEFAULT_LLM_CONFIG,
            "contextual_response_rephraser.init",
            ContextualResponseRephraser.__name__,
        )

    @classmethod
    def _add_prompt_and_llm_metadata_to_response(
        cls,
        response: Dict[str, Any],
        prompt_name: str,
        user_prompt: str,
        llm_response: Optional["LLMResponse"] = None,
    ) -> Dict[str, Any]:
        """Stores the prompt and LLMResponse metadata to response.

        Args:
            response: The response to add the prompt and LLMResponse metadata to.
            prompt_name: A name identifying prompt usage.
            user_prompt: The user prompt that was sent to the LLM.
            llm_response: The response object from the LLM (None if no response).
        """
        from rasa.dialogue_understanding.utils import record_commands_and_prompts

        if not record_commands_and_prompts:
            return response

        prompt_data: Dict[Text, Any] = {
            KEY_COMPONENT_NAME: cls.__name__,
            KEY_PROMPT_NAME: prompt_name,
            KEY_USER_PROMPT: user_prompt,
            KEY_LLM_RESPONSE_METADATA: llm_response.to_dict() if llm_response else None,
        }

        prompts = response.get(PROMPTS, [])
        prompts.append(prompt_data)
        response[PROMPTS] = prompts
        return response

    @staticmethod
    def get_language_label(tracker: DialogueStateTracker) -> str:
        """Fetches the label of the language to be used for the rephraser.

        Args:
            tracker: The tracker to get the language from.

        Returns:
            The label of the current language, or "English" if no language is set.
        """
        return (
            tracker.current_language.label
            if tracker.current_language
            else tracker.default_language.label
        )

    def _last_message_if_human(self, tracker: DialogueStateTracker) -> Optional[str]:
        """Returns the latest message from the tracker.

        If the latest message is from the AI, it returns None.

        Args:
            tracker: The tracker to get the latest message from.

        Returns:
            The latest message from the tracker if it is from the user, else None.
        """
        for event in reversed(tracker.events):
            if isinstance(event, UserUttered):
                return event.text
            if isinstance(event, BotUttered):
                return None
        return None

    @measure_llm_latency
    async def _generate_llm_response(self, prompt: str) -> Optional[LLMResponse]:
        """Use LLM to generate a response.

        Returns an LLMResponse object containing both the generated text
        (choices) and metadata.

        Args:
            prompt: The prompt to send to the LLM.

        Returns:
            An LLMResponse object if successful, otherwise None.
        """
        llm = llm_factory(self.llm_config, DEFAULT_LLM_CONFIG)

        try:
            return await llm.acompletion(prompt)
        except Exception as e:
            # unfortunately, langchain does not wrap LLM exceptions which means
            # we have to catch all exceptions here
            structlogger.error("nlg.llm.error", error=e)
            return None

    def llm_property(self, prop: str) -> Optional[str]:
        """Returns a property of the LLM provider."""
        return combine_custom_and_default_config(
            self.llm_config, DEFAULT_LLM_CONFIG
        ).get(prop)

    def custom_prompt_template(self, prompt_template: str) -> Optional[str]:
        """Returns the custom prompt template if it is not the default one."""
        if prompt_template != DEFAULT_RESPONSE_VARIATION_PROMPT_TEMPLATE:
            return prompt_template
        else:
            return None

    def _template_for_response_rephrasing(self, response: Dict[str, Any]) -> str:
        """Returns the template for the response rephrasing.

        Args:
            response: The response to rephrase.

        Returns:
            The template for the response rephrasing.
        """
        return response.get("metadata", {}).get(
            RESPONSE_REPHRASING_TEMPLATE_KEY, self.prompt_template
        )

    async def _create_history(self, tracker: DialogueStateTracker) -> str:
        """Creates the history for the prompt.

        Args:
            tracker: The tracker to use for the history.


        Returns:
        The history for the prompt.
        """
        # Count multiple utterances by bot/user as single turn in conversation history
        turns_wrapper = (
            _count_multiple_utterances_as_single_turn
            if self.count_multiple_utterances_as_single_turn
            else None
        )
        llm = llm_factory(self.llm_config, DEFAULT_LLM_CONFIG)
        return await summarize_conversation(
            tracker, llm, max_turns=5, turns_wrapper=turns_wrapper
        )

    async def rephrase(
        self,
        response: Dict[str, Any],
        tracker: DialogueStateTracker,
    ) -> Dict[str, Any]:
        """Predicts a variation of the response.

        Args:
            response: The response to rephrase.
            tracker: The tracker to use for the prediction.
            model_name: The name of the model to use for the prediction.

        Returns:
            The response with the rephrased text.
        """
        if not (response_text := response.get(KEY_RESPONSES_TEXT)):
            return response

        prompt_template_text = self._template_for_response_rephrasing(response)

        # Last user message (=current input) should always be in prompt if available
        last_message_by_user = getattr(tracker.latest_message, "text", "")
        current_input = (
            f"{USER}: {last_message_by_user}" if last_message_by_user else ""
        )

        # Only summarise conversation history if flagged
        if self.summarize_history:
            history = await self._create_history(tracker)
        else:
            # Count multiple utterances by bot/user as single turn
            turns_wrapper = (
                _count_multiple_utterances_as_single_turn
                if self.count_multiple_utterances_as_single_turn
                else None
            )
            max_turns = max(self.max_historical_turns, 1)
            history = tracker_as_readable_transcript(
                tracker, max_turns=max_turns, turns_wrapper=turns_wrapper
            )

        prompt = Template(prompt_template_text).render(
            history=history,
            suggested_response=response_text,
            current_input=current_input,
            slots=tracker.current_slot_values(),
            language=self.get_language_label(tracker),
        )
        log_llm(
            logger=structlogger,
            log_module="ContextualResponseRephraser",
            log_event="nlg.rephrase.prompt_rendered",
            prompt=prompt,
        )
        telemetry.track_response_rephrase(
            rephrase_all=self.rephrase_all,
            custom_prompt_template=self.custom_prompt_template(prompt_template_text),
            llm_type=self.llm_property(PROVIDER_CONFIG_KEY),
            llm_model=self.llm_property(MODEL_CONFIG_KEY)
            or self.llm_property(MODEL_NAME_CONFIG_KEY),
            llm_model_group_id=self.llm_property(MODEL_GROUP_ID_CONFIG_KEY),
        )
        llm_response = await self._generate_llm_response(prompt)
        llm_response = LLMResponse.ensure_llm_response(llm_response)

        response = self._add_prompt_and_llm_metadata_to_response(
            response=response,
            prompt_name="rephrase_prompt",
            user_prompt=prompt,
            llm_response=llm_response,
        )

        if not (llm_response and llm_response.choices and llm_response.choices[0]):
            # If the LLM fails to generate a response, return the original response.
            return response

        updated_text = llm_response.choices[0]
        structlogger.debug(
            "nlg.rewrite.complete",
            response_text=response_text,
            updated_text=updated_text,
        )
        response[KEY_RESPONSES_TEXT] = updated_text
        return response

    def does_response_allow_rephrasing(self, template: Dict[Text, Any]) -> bool:
        """Checks if the template allows variation.

        Args:
            template: The template to check.

        Returns:
            `True` if the template allows variation, else `False`.
        """
        return template.get("metadata", {}).get(
            RESPONSE_REPHRASING_KEY, self.rephrase_all
        )

    async def generate(
        self,
        utter_action: Text,
        tracker: DialogueStateTracker,
        output_channel: Text,
        **kwargs: Any,
    ) -> Optional[Dict[Text, Any]]:
        """Generate a response for the requested utter action.

        Args:
            utter_action: The name of the utter action to generate a response for.
            tracker: The tracker to use for the generation.
            output_channel: The output channel to use for the generation.
            **kwargs: Additional arguments to pass to the generation.

        Returns:
            The generated response.
        """
        templated_response = await super().generate(
            utter_action=utter_action,
            tracker=tracker,
            output_channel=output_channel,
            **kwargs,
        )

        if templated_response and self.does_response_allow_rephrasing(
            templated_response
        ):
            return await self.rephrase(
                templated_response,
                tracker,
            )
        else:
            return templated_response



import os
import shutil
import tempfile
import uuid
from enum import Enum
from pathlib import Path
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Iterable,
    List,
    Optional,
    Protocol,
    Set,
    Text,
    Union,
    runtime_checkable,
)

import structlog

if TYPE_CHECKING:
    from rasa.core.channels import UserMessage
    from rasa.shared.core.domain import Domain
    from rasa.shared.nlu.training_data.message import Message

YAML_FILE_EXTENSIONS = [".yml", ".yaml"]
JSON_FILE_EXTENSIONS = [".json"]
TRAINING_DATA_EXTENSIONS = set(JSON_FILE_EXTENSIONS + YAML_FILE_EXTENSIONS)

structlogger = structlog.get_logger()


def yaml_file_extension() -> Text:
    """Return YAML file extension."""
    return YAML_FILE_EXTENSIONS[0]


def is_likely_yaml_file(file_path: Union[Text, Path]) -> bool:
    """Check if a file likely contains yaml.

    Arguments:
        file_path: path to the file

    Returns:
        `True` if the file likely contains data in yaml format, `False` otherwise.
    """
    return Path(file_path).suffix in set(YAML_FILE_EXTENSIONS)


def is_likely_json_file(file_path: Text) -> bool:
    """Check if a file likely contains json.

    Arguments:
        file_path: path to the file

    Returns:
        `True` if the file likely contains data in json format, `False` otherwise.
    """
    return Path(file_path).suffix in set(JSON_FILE_EXTENSIONS)


def get_core_directory(paths: Optional[Union[Text, List[Text]]]) -> Text:
    """Recursively collects all Core training files from a list of paths.

    Args:
        paths: List of paths to training files or folders containing them.

    Returns:
        Path to temporary directory containing all found Core training files.
    """
    from rasa.shared.core.training_data.story_reader.yaml_story_reader import (
        YAMLStoryReader,
    )

    core_files = get_data_files(paths, YAMLStoryReader.is_stories_file)
    return _copy_files_to_new_dir(core_files)


def get_nlu_directory(paths: Optional[Union[Text, List[Text]]]) -> Text:
    """Recursively collects all NLU training files from a list of paths.

    Args:
        paths: List of paths to training files or folders containing them.

    Returns:
        Path to temporary directory containing all found NLU training files.
    """
    nlu_files = get_data_files(paths, is_nlu_file)
    return _copy_files_to_new_dir(nlu_files)


def get_data_files(
    paths: Optional[Union[Text, List[Text]]], filter_predicate: Callable[[Text], bool]
) -> List[Text]:
    """Recursively collects all training files from a list of paths.

    Args:
        paths: List of paths to training files or folders containing them.
        filter_predicate: property to use when filtering the paths, e.g. `is_nlu_file`.

    Returns:
        Paths of training data files.
    """
    data_files = set()

    if paths is None:
        paths = []
    elif isinstance(paths, str):
        paths = [paths]

    for path in set(paths):
        if not path:
            continue

        if is_valid_filetype(path):
            if filter_predicate(path):
                data_files.add(os.path.abspath(path))
        else:
            new_data_files = _find_data_files_in_directory(path, filter_predicate)
            data_files.update(new_data_files)

    return sorted(data_files)


def _find_data_files_in_directory(
    directory: Text, filter_property: Callable[[Text], bool]
) -> Set[Text]:
    filtered_files = set()

    for root, _, files in os.walk(directory, followlinks=True):
        # we sort the files here to ensure consistent order for repeatable training
        # results
        for f in sorted(files):
            full_path = os.path.join(root, f)

            if not is_valid_filetype(full_path):
                continue

            if filter_property(full_path):
                filtered_files.add(full_path)

    return filtered_files


def is_valid_filetype(path: Union[Path, Text]) -> bool:
    """Checks if given file has a supported extension.

    Args:
        path: Path to the source file.

    Returns:
        `True` is given file has supported extension, `False` otherwise.
    """
    return Path(path).is_file() and Path(path).suffix in TRAINING_DATA_EXTENSIONS


def is_nlu_file(file_path: Text) -> bool:
    """Checks if a file is a Rasa compatible nlu file.

    Args:
        file_path: Path of the file which should be checked.

    Returns:
        `True` if it's a nlu file, otherwise `False`.
    """
    from rasa.shared.nlu.training_data import loading as nlu_loading

    return nlu_loading.guess_format(file_path) != nlu_loading.UNK


def is_config_file(file_path: Text) -> bool:
    """Checks whether the given file path is a Rasa config file.

    Args:
        file_path: Path of the file which should be checked.

    Returns:
        `True` if it's a Rasa config file, otherwise `False`.
    """
    file_name = os.path.basename(file_path)

    return file_name in ["config.yml", "config.yaml"]


def _copy_files_to_new_dir(files: Iterable[Text]) -> Text:
    directory = tempfile.mkdtemp()
    for f in files:
        # makes sure files do not overwrite each other, hence the prefix
        unique_prefix = uuid.uuid4().hex
        unique_file_name = unique_prefix + "_" + os.path.basename(f)
        shutil.copy2(f, os.path.join(directory, unique_file_name))

    return directory


class TrainingType(Enum):
    """Enum class for defining explicitly what training types exist."""

    NLU = 1
    CORE = 2
    BOTH = 3
    END_TO_END = 4

    @property
    def model_type(self) -> Text:
        """Returns the type of model which this training yields."""
        if self == TrainingType.NLU:
            return "nlu"
        if self == TrainingType.CORE:
            return "core"
        return "rasa"


@runtime_checkable
class RegexReader(Protocol):
    def unpack_regex_message(self, message: "Message", **kwargs: Any) -> "Message": ...


def create_regex_pattern_reader(
    message: "UserMessage", domain: "Domain"
) -> Optional[RegexReader]:
    """Create a new instance of a class to unpack regex patterns.

    Args:
        message: The user message to unpack.
        domain: The domain of the assistant.
    """
    from rasa.shared.core.command_payload_reader import CommandPayloadReader
    from rasa.shared.core.training_data.story_reader.yaml_story_reader import (
        YAMLStoryReader,
    )

    if message.text is None:
        structlogger.warning(
            "empty.user.message",
            user_text="None",
        )
        return None

    reader: Union[CommandPayloadReader, YAMLStoryReader, None] = None

    if message.text.startswith("/SetSlots"):
        reader = CommandPayloadReader()
    elif message.text.startswith("/"):
        reader = YAMLStoryReader(domain)

    return reader if isinstance(reader, RegexReader) else None



from __future__ import annotations

from dataclasses import dataclass
from typing import Any, Dict, List

from rasa.dialogue_understanding.commands.command import Command
from rasa.shared.core.events import Event
from rasa.shared.core.flows import FlowsList
from rasa.shared.core.trackers import DialogueStateTracker


@dataclass
class NoopCommand(Command):
    """A command to indicate that nothing needs to be done."""

    @classmethod
    def command(cls) -> str:
        """Returns the command type."""
        return "noop"

    @classmethod
    def from_dict(cls, data: Dict[str, Any]) -> NoopCommand:
        """Converts the dictionary to a command.

        Returns:
            The converted dictionary.
        """
        return NoopCommand()

    def run_command_on_tracker(
        self,
        tracker: DialogueStateTracker,
        all_flows: FlowsList,
        original_tracker: DialogueStateTracker,
    ) -> List[Event]:
        """Runs the command on the tracker.

        Args:
            tracker: The tracker to run the command on.
            all_flows: All flows in the assistant.
            original_tracker: The tracker before any command was executed.

        Returns:
            The events to apply to the tracker.
        """
        return []

    def __hash__(self) -> int:
        return hash(self.command())

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, NoopCommand):
            return False

        return True



from dataclasses import dataclass
from typing import Any, Dict, List, Optional

from rasa.shared.core.events import Event


@dataclass
class FlowActionPrediction:
    """Represents an action prediction."""

    action_name: Optional[str]
    """The name of the predicted action."""
    score: float
    """The score of the predicted action."""
    metadata: Optional[Dict[str, Any]] = None
    """The metadata of the predicted action."""
    events: Optional[List[Event]] = None
    """The events attached to the predicted action."""


class FlowStepResult:
    """Represents the result of a flow step."""

    def __init__(self, events: Optional[List[Event]] = None) -> None:
        self.events = events or []


class ContinueFlowWithNextStep(FlowStepResult):
    """Represents the result of a flow step that should continue with the next step."""

    def __init__(
        self, events: Optional[List[Event]] = None, has_flow_ended: bool = False
    ) -> None:
        self.has_flow_ended = has_flow_ended
        super().__init__(events=events)


class PauseFlowReturnPrediction(FlowStepResult):
    """Result where the flow execution should be paused after this step."""

    def __init__(self, action_prediction: FlowActionPrediction) -> None:
        self.action_prediction = action_prediction
        super().__init__(events=action_prediction.events)



import abc
from typing import Any, Dict, Type, TypeVar

OAUTH_TYPE_FIELD = "type"
OAUTH_KEY = "oauth"

OAuthType = TypeVar("OAuthType", bound="OAuth")


class OAuth(abc.ABC):
    """Interface for OAuth configuration."""

    @classmethod
    @abc.abstractmethod
    def from_dict(
        cls: Type[OAuthType], config: Dict[str, Any]
    ) -> OAuthType:  # ignore[type]
        """Initializes a dataclass from the passed config.

        Args:
            config: (dict) The config from which to initialize.

        Returns:
            OAuth
        """

    @abc.abstractmethod
    def get_bearer_token(self) -> str:
        """Returns a bearer token.

        Bear token is used to authenticate requests to the Azure
        Oopen AI instance's API protected by the Gateway.
        """



import rasa.shared.nlu.constants
from rasa.shared.nlu.constants import ENTITY_ATTRIBUTE_CONFIDENCE

BILOU_ENTITIES = "bilou_entities"
BILOU_ENTITIES_ROLE = "bilou_entities_role"
BILOU_ENTITIES_GROUP = "bilou_entities_group"

ENTITY_ATTRIBUTE_CONFIDENCE_TYPE = (
    f"{ENTITY_ATTRIBUTE_CONFIDENCE}_{rasa.shared.nlu.constants.ENTITY_ATTRIBUTE_TYPE}"
)
ENTITY_ATTRIBUTE_CONFIDENCE_GROUP = (
    f"{ENTITY_ATTRIBUTE_CONFIDENCE}_{rasa.shared.nlu.constants.ENTITY_ATTRIBUTE_GROUP}"
)
ENTITY_ATTRIBUTE_CONFIDENCE_ROLE = (
    f"{ENTITY_ATTRIBUTE_CONFIDENCE}_{rasa.shared.nlu.constants.ENTITY_ATTRIBUTE_ROLE}"
)

EXTRACTOR = "extractor"

PRETRAINED_EXTRACTORS = {"DucklingEntityExtractor", "SpacyEntityExtractor"}

NUMBER_OF_SUB_TOKENS = "number_of_sub_tokens"

MESSAGE_ATTRIBUTES = [
    rasa.shared.nlu.constants.TEXT,
    rasa.shared.nlu.constants.INTENT,
    rasa.shared.nlu.constants.RESPONSE,
    rasa.shared.nlu.constants.ACTION_NAME,
    rasa.shared.nlu.constants.ACTION_TEXT,
    rasa.shared.nlu.constants.INTENT_RESPONSE_KEY,
]
# the dense featurizable attributes are essentially text attributes
DENSE_FEATURIZABLE_ATTRIBUTES = [
    rasa.shared.nlu.constants.TEXT,
    rasa.shared.nlu.constants.RESPONSE,
    rasa.shared.nlu.constants.ACTION_TEXT,
]

LANGUAGE_MODEL_DOCS = {
    rasa.shared.nlu.constants.TEXT: "text_language_model_doc",
    rasa.shared.nlu.constants.RESPONSE: "response_language_model_doc",
    rasa.shared.nlu.constants.ACTION_TEXT: "action_text_model_doc",
}
SPACY_DOCS = {
    rasa.shared.nlu.constants.TEXT: "text_spacy_doc",
    rasa.shared.nlu.constants.RESPONSE: "response_spacy_doc",
    rasa.shared.nlu.constants.ACTION_TEXT: "action_text_spacy_doc",
}

TOKENS_NAMES = {
    rasa.shared.nlu.constants.TEXT: "text_tokens",
    rasa.shared.nlu.constants.INTENT: "intent_tokens",
    rasa.shared.nlu.constants.RESPONSE: "response_tokens",
    rasa.shared.nlu.constants.ACTION_NAME: "action_name_tokens",
    rasa.shared.nlu.constants.ACTION_TEXT: "action_text_tokens",
    rasa.shared.nlu.constants.INTENT_RESPONSE_KEY: "intent_response_key_tokens",
}

SEQUENCE_FEATURES = "sequence_features"
SENTENCE_FEATURES = "sentence_features"

RESPONSE_SELECTOR_PROPERTY_NAME = "response_selector"
RESPONSE_SELECTOR_RETRIEVAL_INTENTS = "all_retrieval_intents"
RESPONSE_SELECTOR_DEFAULT_INTENT = "default"
RESPONSE_SELECTOR_PREDICTION_KEY = "response"
RESPONSE_SELECTOR_RANKING_KEY = "ranking"
RESPONSE_SELECTOR_RESPONSES_KEY = "responses"
RESPONSE_SELECTOR_RESPONSE_TEMPLATES_KEY = "response_templates"
RESPONSE_SELECTOR_UTTER_ACTION_KEY = "utter_action"
RESPONSE_SELECTOR_TEMPLATE_NAME_KEY = "template_name"
RESPONSE_IDENTIFIER_DELIMITER = "/"

DEFAULT_TRANSFORMER_SIZE = 256

FEATURIZER_CLASS_ALIAS = "alias"

NO_LENGTH_RESTRICTION = -1



import os
import shutil
import subprocess
from enum import Enum
from pathlib import Path
from typing import Any, Dict, Optional

import structlog
from pydantic import BaseModel, ConfigDict

from rasa.constants import MODEL_ARCHIVE_EXTENSION, RASA_DIR_NAME
from rasa.model_manager import config
from rasa.model_manager.utils import (
    ensure_base_directory_exists,
    logs_path,
    models_base_path,
    subpath,
    write_encoded_data_to_file,
)
from rasa.model_manager.warm_rasa_process import (
    start_rasa_process,
)
from rasa.model_training import generate_random_model_name
from rasa.studio.prompts import handle_prompts

structlogger = structlog.get_logger()


class TrainingSessionStatus(str, Enum):
    """Enum for the training status."""

    RUNNING = "running"
    STOPPED = "stopped"
    DONE = "done"
    ERROR = "error"


class TrainingSession(BaseModel):
    """Store information about a training session."""

    model_config = ConfigDict(arbitrary_types_allowed=True)

    training_id: str
    assistant_id: str
    client_id: Optional[str]
    progress: int
    model_name: str
    status: TrainingSessionStatus
    process: subprocess.Popen
    log_id: str

    def is_status_indicating_alive(self) -> bool:
        """Check if the training is running."""
        return self.status == TrainingSessionStatus.RUNNING

    def model_path(self) -> str:
        """Return the path to the model."""
        return subpath(models_base_path(), f"{self.model_name}.tar.gz")


def train_path(training_id: str) -> str:
    """Return the path to the training directory for a given training id."""
    return subpath(config.SERVER_BASE_WORKING_DIRECTORY + "/trainings", training_id)


def cache_for_assistant_path(assistant_id: str) -> str:
    """Return the path to the cache directory for a given assistant id."""
    return subpath(config.SERVER_BASE_WORKING_DIRECTORY + "/caches", assistant_id)


def terminate_training(training: TrainingSession) -> None:
    if not training.is_status_indicating_alive():
        # no-op if the training is not running
        return

    structlogger.info(
        "model_trainer.user_stopping_training", training_id=training.training_id
    )
    try:
        training.process.terminate()
        training.status = TrainingSessionStatus.STOPPED
    except ProcessLookupError:
        structlogger.debug(
            "model_trainer.training_process_not_found",
            training_id=training.training_id,
        )
    finally:
        clean_up_after_training(training)


def update_training_status(training: TrainingSession) -> None:
    if not training.is_status_indicating_alive():
        # skip if the training is not running
        return
    if training.process.poll() is None:
        # process is still running
        return

    complete_training(training)


def complete_training(training: TrainingSession) -> None:
    """Complete a training session.

    Transitions the status of a training process to "done" if the process has
    finished successfully, and to "error" if the process has finished with an
    error.
    """
    if training.process.returncode == 0:
        training.status = TrainingSessionStatus.DONE
    else:
        training.status = TrainingSessionStatus.ERROR

    training.progress = 100

    structlogger.info(
        "model_trainer.training_finished",
        training_id=training.training_id,
        status=training.status,
    )

    # persist the assistant cache to speed up future training runs for this
    # assistant
    persist_rasa_cache(training.assistant_id, train_path(training.training_id))
    move_model_to_local_storage(training)

    clean_up_after_training(training)


def clean_up_after_training(training: TrainingSession) -> None:
    """Clean up the training directory."""
    structlogger.debug(
        "model_trainer.cleaning_up_training", training_id=training.training_id
    )
    shutil.rmtree(train_path(training.training_id), ignore_errors=True)


def move_model_to_local_storage(training: TrainingSession) -> None:
    """Persist the model to the remote storage."""
    ensure_base_directory_exists(models_base_path())

    model_path = subpath(
        train_path(training.training_id) + "/models",
        f"{training.model_name}.{MODEL_ARCHIVE_EXTENSION}",
    )

    if os.path.exists(model_path):
        structlogger.debug(
            "model_trainer.persisting_model_to_models_dir",
            training_model_path=model_path,
            storage_model_path=models_base_path(),
        )
        shutil.move(model_path, models_base_path())
    else:
        structlogger.warning(
            "model_trainer.model_not_found_after_training",
            training_id=training.training_id,
            model_path=model_path,
        )


def seed_training_directory_with_rasa_cache(
    training_base_path: str, assistant_id: str
) -> None:
    """Populate the training directory with the cache of a previous training."""
    # check if there is a cache for this assistant
    cache_path = cache_for_assistant_path(assistant_id)

    if os.path.exists(cache_path):
        structlogger.debug(
            "model_trainer.populating_training_dir_with_cache",
            assistant_id=assistant_id,
            training_base_path=training_base_path,
        )
        # copy the cache to the training directory
        shutil.copytree(src=cache_path, dst=subpath(training_base_path, RASA_DIR_NAME))


def persist_rasa_cache(assistant_id: str, training_base_path: str) -> None:
    """Persist the cache of a training session to speed up future trainings."""
    # copy the cache from the training directory to the cache directory
    # cache files are stored inside of `/.rasa/` of the training folder
    structlogger.debug(
        "model_trainer.persisting_assistant_cache", assistant_id=assistant_id
    )
    cache_path = cache_for_assistant_path(assistant_id)

    # if the training failed and didn't create a cache, skip this step
    if not os.path.exists(subpath(training_base_path, RASA_DIR_NAME)):
        return

    # clean up the cache directory first
    shutil.rmtree(cache_path, ignore_errors=True)
    shutil.copytree(src=subpath(training_base_path, RASA_DIR_NAME), dst=cache_path)


def write_training_data_to_files(
    encoded_training_data: Dict[str, Any], training_base_path: str
) -> None:
    """Write the training data to files in the training directory.

    Incoming data format, all keys being optional:
    ````
    {
        "domain": "base64 encoded domain.yml",
        "credentials": "base64 encoded credentials.yml",
        "endpoints": "base64 encoded endpoints.yml",
        "flows": "base64 encoded flows.yml",
        "config": "base64 encoded config.yml",
        "stories": "base64 encoded stories.yml",
        "rules": "base64 encoded rules.yml",
        "nlu": "base64 encoded nlu.yml"
        "prompts": "dictionary with the prompts",
    }
    ```
    """
    data_to_be_written_to_files = {
        "domain": "domain.yml",
        "credentials": "credentials.yml",
        "endpoints": "endpoints.yml",
        "flows": "data/flows.yml",
        "config": "config.yml",
        "stories": "data/stories.yml",
        "rules": "data/rules.yml",
        "nlu": "data/nlu.yml",
    }

    for key, file in data_to_be_written_to_files.items():
        parent_path, file_name = os.path.split(file)

        write_encoded_data_to_file(
            encoded_training_data.get(key, ""),
            subpath(training_base_path + "/" + parent_path, file_name),
        )

    if prompts := encoded_training_data.get("prompts"):
        handle_prompts(prompts, Path(training_base_path))


def prepare_training_directory(
    training_base_path: str, assistant_id: str, encoded_training_data: Dict[str, Any]
) -> None:
    """Prepare the training directory for a new training session."""
    # create a new working directory and store the training data from the
    # request there. the training data in the request is base64 encoded
    os.makedirs(training_base_path, exist_ok=True)

    seed_training_directory_with_rasa_cache(training_base_path, assistant_id)
    write_training_data_to_files(encoded_training_data, training_base_path)
    structlogger.debug("model_trainer.prepared_training", path=training_base_path)


def start_training_process(
    training_id: str,
    assistant_id: str,
    client_id: str,
    training_base_path: str,
) -> TrainingSession:
    model_name = generate_random_model_name()
    # Start the training in a subprocess
    # set the working directory to the training directory
    # run the rasa train command as a subprocess, activating poetry before running
    # pipe the stdout and stderr to the same file
    arguments = [
        "train",
        "--debug",
        "--data",
        "data",
        "--config",
        "config.yml",
        "--domain",
        "domain.yml",
        "--endpoints",
        "endpoints.yml",
        "--fixed-model-name",
        f"{model_name}.{MODEL_ARCHIVE_EXTENSION}",
        "--out",
        "models",
    ]

    if config.SERVER_MODEL_REMOTE_STORAGE:
        arguments.extend(
            [
                "--keep-local-model-copy",
                "--remote-storage",
                config.SERVER_MODEL_REMOTE_STORAGE,
                "--remote-root-only",
            ]
        )

    structlogger.debug(
        "model_trainer.training_arguments", arguments=" ".join(arguments)
    )

    warm_process = start_rasa_process(cwd=training_base_path, arguments=arguments)

    structlogger.info(
        "model_trainer.training_started",
        training_id=training_id,
        assistant_id=assistant_id,
        model_name=model_name,
        client_id=client_id,
        log=logs_path(warm_process.log_id),
        pid=warm_process.process.pid,
    )

    return TrainingSession(
        training_id=training_id,
        assistant_id=assistant_id,
        client_id=client_id,
        model_name=model_name,
        progress=0,
        status=TrainingSessionStatus.RUNNING,
        process=warm_process.process,  # Store the process handle
        log_id=warm_process.log_id,
    )


def run_training(
    training_id: str,
    assistant_id: str,
    client_id: str,
    encoded_training_data: Dict,
) -> TrainingSession:
    """Run a training session."""
    training_base_path = train_path(training_id)

    prepare_training_directory(training_base_path, assistant_id, encoded_training_data)
    return start_training_process(
        training_id=training_id,
        assistant_id=assistant_id,
        client_id=client_id,
        training_base_path=training_base_path,
    )



from __future__ import annotations

from dataclasses import dataclass
from typing import Any, Dict, List, Optional, Text

from rasa.shared.core.flows.flow_step import FlowStep, step_from_json
from rasa.shared.core.flows.steps.internal import InternalFlowStep


@dataclass
class FlowStepSequence:
    """A Sequence of flow steps."""

    child_steps: List[FlowStep]

    @staticmethod
    def from_json(flow_id: Text, data: List[Dict[Text, Any]]) -> FlowStepSequence:
        """Create a FlowStepSequence object from serialized data

        Args:
            flow_id: The id of the flow that contains the step.
            data: data for a StepSequence in a serialized format

        Returns:
            A StepSequence object including its flow step objects.
        """
        flow_steps: List[FlowStep] = [
            step_from_json(flow_id, config) for config in data
        ]

        return FlowStepSequence(child_steps=flow_steps)

    def as_json(self) -> List[Dict[Text, Any]]:
        """Serialize the StepSequence object and contained FlowStep objects

        Returns:
            the FlowStepSequence and its FlowSteps as serialized data
        """
        return [
            step.as_json()
            for step in self.child_steps
            if not isinstance(step, InternalFlowStep)
        ]

    def _resolve_steps(self, should_resolve_calls: bool) -> List[FlowStep]:
        """Resolves the steps of the flow."""
        return [
            step
            for child_step in self.child_steps
            for step in child_step.steps_in_tree(
                should_resolve_calls=should_resolve_calls
            )
        ]

    @property
    def steps_with_calls_resolved(self) -> List[FlowStep]:
        """Return all steps in this step sequence and their sub steps."""
        return self._resolve_steps(should_resolve_calls=True)

    @property
    def steps(self) -> List[FlowStep]:
        """Return the steps of the flow without steps of called flows"""
        return self._resolve_steps(should_resolve_calls=False)

    def first(self) -> Optional[FlowStep]:
        """Return the first step of the sequence."""
        return self.child_steps[0] if self.child_steps else None

    @classmethod
    def empty(cls) -> FlowStepSequence:
        """Create an empty FlowStepSequence object."""
        return cls(child_steps=[])

    def __eq__(self, other: object) -> bool:
        return (
            isinstance(other, FlowStepSequence)
            and self.child_steps == other.child_steps
        )



from __future__ import annotations

from typing import Any, Dict, List, Optional, Text

import structlog
from jinja2 import Template
from pypred import Predicate
from structlog.contextvars import (
    bound_contextvars,
)

from rasa.core.available_endpoints import AvailableEndpoints
from rasa.core.constants import ACTIVE_FLOW_METADATA_KEY, STEP_ID_METADATA_KEY
from rasa.core.policies.flows.flow_exceptions import (
    FlowCircuitBreakerTrippedException,
    FlowException,
    NoNextStepInFlowException,
)
from rasa.core.policies.flows.flow_step_result import (
    ContinueFlowWithNextStep,
    FlowActionPrediction,
    FlowStepResult,
    PauseFlowReturnPrediction,
)
from rasa.dialogue_understanding.commands import CancelFlowCommand
from rasa.dialogue_understanding.patterns.cancel import CancelPatternFlowStackFrame
from rasa.dialogue_understanding.patterns.collect_information import (
    FLOW_PATTERN_COLLECT_INFORMATION,
    CollectInformationPatternFlowStackFrame,
)
from rasa.dialogue_understanding.patterns.completed import (
    CompletedPatternFlowStackFrame,
)
from rasa.dialogue_understanding.patterns.continue_interrupted import (
    ContinueInterruptedPatternFlowStackFrame,
)
from rasa.dialogue_understanding.patterns.human_handoff import (
    HumanHandoffPatternFlowStackFrame,
)
from rasa.dialogue_understanding.patterns.internal_error import (
    InternalErrorPatternFlowStackFrame,
)
from rasa.dialogue_understanding.patterns.search import SearchPatternFlowStackFrame
from rasa.dialogue_understanding.patterns.user_silence import FLOW_PATTERN_USER_SILENCE
from rasa.dialogue_understanding.stack.dialogue_stack import DialogueStack
from rasa.dialogue_understanding.stack.frames import (
    BaseFlowStackFrame,
    DialogueStackFrame,
    UserFlowStackFrame,
)
from rasa.dialogue_understanding.stack.frames.flow_stack_frame import (
    FlowStackFrameType,
)
from rasa.dialogue_understanding.stack.utils import (
    top_user_flow_frame,
)
from rasa.shared.constants import RASA_PATTERN_HUMAN_HANDOFF
from rasa.shared.core.constants import (
    ACTION_LISTEN_NAME,
    SILENCE_TIMEOUT_SLOT,
)
from rasa.shared.core.events import (
    Event,
    FlowCompleted,
    FlowResumed,
    FlowStarted,
    SlotSet,
)
from rasa.shared.core.flows import FlowsList
from rasa.shared.core.flows.flow import (
    END_STEP,
    Flow,
    FlowStep,
)
from rasa.shared.core.flows.flow_step_links import (
    ElseFlowStepLink,
    IfFlowStepLink,
    StaticFlowStepLink,
)
from rasa.shared.core.flows.steps import (
    ActionFlowStep,
    CallFlowStep,
    CollectInformationFlowStep,
    ContinueFlowStep,
    EndFlowStep,
    LinkFlowStep,
    NoOperationFlowStep,
    SetSlotsFlowStep,
)
from rasa.shared.core.flows.steps.constants import START_STEP
from rasa.shared.core.slots import Slot, SlotRejection
from rasa.shared.core.trackers import (
    DialogueStateTracker,
)

structlogger = structlog.get_logger()

MAX_NUMBER_OF_STEPS = 250


def render_template_variables(text: str, context: Dict[Text, Any]) -> str:
    """Replace context variables in a text."""
    return Template(text).render(context)


def is_condition_satisfied(
    predicate: Text, context: Dict[str, Any], tracker: DialogueStateTracker
) -> bool:
    """Evaluate a predicate condition."""
    # attach context to the predicate evaluation to allow conditions using it
    context = {"context": context}

    document: Dict[str, Any] = context.copy()
    # add slots namespace to the document
    document["slots"] = tracker.current_slot_values()

    rendered_condition = render_template_variables(predicate, context)
    p = Predicate(rendered_condition)
    structlogger.debug(
        "flow.predicate.evaluating",
        condition=predicate,
        rendered_condition=rendered_condition,
    )
    try:
        return p.evaluate(document)
    except (TypeError, Exception) as e:
        structlogger.error(
            "flow.predicate.error",
            predicate=predicate,
            error=str(e),
        )
        return False


def is_step_end_of_flow(step: FlowStep) -> bool:
    """Check if a step is the end of a flow."""
    return (
        step.id == END_STEP
        or
        # not quite at the end but almost, so we'll treat it as the end
        step.id == ContinueFlowStep.continue_step_for_id(END_STEP)
    )


def select_next_step_id(
    current: FlowStep,
    condition_evaluation_context: Dict[str, Any],
    tracker: DialogueStateTracker,
) -> Optional[Text]:
    """Selects the next step id based on the current step."""
    next_step = current.next
    if len(next_step.links) == 1 and isinstance(next_step.links[0], StaticFlowStepLink):
        return next_step.links[0].target

    # evaluate if conditions
    for link in next_step.links:
        if isinstance(link, IfFlowStepLink) and link.condition:
            if is_condition_satisfied(
                link.condition, condition_evaluation_context, tracker
            ):
                structlogger.debug(
                    "flow.link.if_condition_satisfied",
                    current_id=current.id,
                    target=link.target,
                )
                return link.target

    # evaluate else condition
    for link in next_step.links:
        if isinstance(link, ElseFlowStepLink):
            structlogger.debug(
                "flow.link.else_condition_satisfied",
                current_id=current.id,
                target=link.target,
            )
            return link.target

    if next_step.links:
        structlogger.error(
            "flow.link.failed_to_select_branch",
            current=current,
            links=next_step.links,
            sender_id=tracker.sender_id,
        )
        return None

    if current.id == END_STEP:
        # we are already at the very end of the flow. There is no next step.
        return None
    elif isinstance(current, LinkFlowStep):
        # link steps don't have a next step, so we'll return the end step
        return END_STEP
    else:
        structlogger.error(
            "flow.step.failed_to_select_next_step",
            step=current,
            sender_id=tracker.sender_id,
        )
        return None


def select_next_step(
    current_step: FlowStep,
    current_flow: Flow,
    stack: DialogueStack,
    tracker: DialogueStateTracker,
) -> Optional[FlowStep]:
    """Get the next step to execute."""
    next_id = select_next_step_id(current_step, stack.current_context(), tracker)
    step = current_flow.step_by_id(next_id)
    structlogger.debug(
        "flow.step.next",
        next_id=step.id if step else None,
        current_id=current_step.id,
        flow_id=current_flow.id,
    )
    return step


def update_top_flow_step_id(updated_id: str, stack: DialogueStack) -> DialogueStack:
    """Update the top flow on the stack."""
    if (top := stack.top()) and isinstance(top, BaseFlowStackFrame):
        top.step_id = updated_id
    return stack


def events_from_set_slots_step(step: SetSlotsFlowStep) -> List[Event]:
    """Create events from a set slots step."""
    return [SlotSet(slot["key"], slot["value"]) for slot in step.slots]


def trigger_pattern_continue_interrupted(
    current_frame: DialogueStackFrame,
    stack: DialogueStack,
    flows: FlowsList,
    tracker: DialogueStateTracker,
) -> List[Event]:
    """Trigger the pattern to continue an interrupted flow if needed."""
    events: List[Event] = []

    # get previously started user flow that will be continued
    interrupted_user_flow_frame = top_user_flow_frame(stack)
    interrupted_user_flow_step = (
        interrupted_user_flow_frame.step(flows) if interrupted_user_flow_frame else None
    )
    interrupted_user_flow = (
        interrupted_user_flow_frame.flow(flows) if interrupted_user_flow_frame else None
    )

    if (
        isinstance(current_frame, UserFlowStackFrame)
        and interrupted_user_flow_step is not None
        and interrupted_user_flow is not None
        and current_frame.frame_type == FlowStackFrameType.INTERRUPT
        and not is_step_end_of_flow(interrupted_user_flow_step)
    ):
        stack.push(
            ContinueInterruptedPatternFlowStackFrame(
                previous_flow_name=interrupted_user_flow.readable_name(
                    language=tracker.current_language
                ),
            )
        )
        events.append(
            FlowResumed(interrupted_user_flow.id, interrupted_user_flow_step.id)
        )

    return events


def trigger_pattern_completed(
    current_frame: DialogueStackFrame, stack: DialogueStack, flows: FlowsList
) -> None:
    """Trigger the pattern indicating that the stack is empty, if needed."""
    # trigger pattern if the stack is empty and the last frame was either a user flow
    # frame or a search frame
    if stack.is_empty() and (
        isinstance(current_frame, UserFlowStackFrame)
        or isinstance(current_frame, SearchPatternFlowStackFrame)
    ):
        completed_flow = current_frame.flow(flows)
        if not completed_flow.run_pattern_completed:
            return

        completed_flow_name = completed_flow.readable_name() if completed_flow else None
        stack.push(
            CompletedPatternFlowStackFrame(
                previous_flow_name=completed_flow_name,
            )
        )


def trigger_pattern_ask_collect_information(
    collect: str,
    stack: DialogueStack,
    rejections: List[SlotRejection],
    utter: str,
    collect_action: str,
) -> None:
    """Trigger the pattern to ask for a slot value."""
    stack.push(
        CollectInformationPatternFlowStackFrame(
            collect=collect,
            utter=utter,
            collect_action=collect_action,
            rejections=rejections,
        )
    )


def reset_scoped_slots(
    current_frame: DialogueStackFrame, current_flow: Flow, tracker: DialogueStateTracker
) -> List[Event]:
    """Reset all scoped slots."""

    def _reset_slot(slot_name: Text, dialogue_tracker: DialogueStateTracker) -> None:
        slot = dialogue_tracker.slots.get(slot_name, None)
        initial_value = slot.initial_value if slot else None
        events.append(SlotSet(slot_name, initial_value, metadata={"reset": True}))

    if (
        isinstance(current_frame, UserFlowStackFrame)
        and current_frame.frame_type == FlowStackFrameType.CALL
    ):
        # if a called frame is completed, we don't reset the slots
        # as they are scoped to the called flow. resetting will happen as part
        # of the flow that contained the call step triggering this called flow
        return []

    events: List[Event] = []

    not_resettable_slot_names = set()
    flow_persistable_slots = current_flow.persisted_slots

    for step in current_flow.steps_with_calls_resolved:
        if isinstance(step, CollectInformationFlowStep):
            # reset all slots scoped to the flow
            slot_name = step.collect
            if step.reset_after_flow_ends and slot_name not in flow_persistable_slots:
                _reset_slot(slot_name, tracker)
            else:
                not_resettable_slot_names.add(slot_name)

    # slots set by the set slots step should be reset after the flow ends
    # unless they are also used in a collect step where `reset_after_flow_ends`
    # is set to `False` or set in the `persisted_slots` list.
    resettable_set_slots = [
        slot["key"]
        for step in current_flow.steps_with_calls_resolved
        if isinstance(step, SetSlotsFlowStep)
        for slot in step.slots
        if slot["key"] not in not_resettable_slot_names
        and slot["key"] not in flow_persistable_slots
    ]

    for name in resettable_set_slots:
        _reset_slot(name, tracker)

    return events


def advance_flows(
    tracker: DialogueStateTracker, available_actions: List[str], flows: FlowsList
) -> FlowActionPrediction:
    """Advance the current flows until the next action.

    Args:
        tracker: The tracker to get the next action for.
        available_actions: The actions that are available in the domain.
        flows: All flows.

    Returns:
    The predicted action and the events to run.
    """
    stack = tracker.stack
    if stack.is_empty():
        # if there are no flows, there is nothing to do
        return FlowActionPrediction(None, 0.0)

    return advance_flows_until_next_action(tracker, available_actions, flows)


def advance_flows_until_next_action(
    tracker: DialogueStateTracker,
    available_actions: List[str],
    flows: FlowsList,
) -> FlowActionPrediction:
    """Advance the flow and select the next action to execute.

    Advances the current flow and returns the next action to execute. A flow
    is advanced until it is completed or until it predicts an action. If
    the flow is completed, the next flow is popped from the stack and
    advanced. If there are no more flows, the action listen is predicted.

    Args:
        tracker: The tracker to get the next action for.
        available_actions: The actions that are available in the domain.
        flows: All flows.

    Returns:
        The next action to execute, the events that should be applied to the
    tracker and the confidence of the prediction.
    """
    step_result: FlowStepResult = ContinueFlowWithNextStep()

    tracker = tracker.copy()

    number_of_initial_events = len(tracker.events)

    number_of_steps_taken = 0

    while isinstance(step_result, ContinueFlowWithNextStep):
        number_of_steps_taken += 1
        if number_of_steps_taken > MAX_NUMBER_OF_STEPS:
            raise FlowCircuitBreakerTrippedException(
                tracker.stack, number_of_steps_taken
            )

        active_frame = tracker.stack.top()
        if not isinstance(active_frame, BaseFlowStackFrame):
            # If there is no current flow, we assume that all flows are done
            # and there is nothing to do. The assumption here is that every
            # flow ends with an action listen.
            step_result = PauseFlowReturnPrediction(
                FlowActionPrediction(ACTION_LISTEN_NAME, 1.0)
            )
            break

        with bound_contextvars(flow_id=active_frame.flow_id):
            previous_step_id = active_frame.step_id
            structlogger.debug("flow.execution.loop", previous_step_id=previous_step_id)
            current_flow = active_frame.flow(flows)
            next_step = select_next_step(
                active_frame.step(flows), current_flow, tracker.stack, tracker
            )

            if not next_step:
                raise NoNextStepInFlowException(tracker.stack)

            tracker.update_stack(update_top_flow_step_id(next_step.id, tracker.stack))

            with bound_contextvars(step_id=next_step.id):
                step_stack = tracker.stack
                step_result = run_step(
                    next_step,
                    current_flow,
                    step_stack,
                    tracker,
                    available_actions,
                    flows,
                    previous_step_id,
                )
                new_events = step_result.events
                if (
                    isinstance(step_result, ContinueFlowWithNextStep)
                    and step_result.has_flow_ended
                ):
                    # insert flow completed before flow resumed event
                    offset = (
                        -1
                        if new_events and isinstance(new_events[-1], FlowResumed)
                        else 0
                    )
                    idx = len(new_events) + offset
                    new_events.insert(
                        idx, FlowCompleted(active_frame.flow_id, previous_step_id)
                    )
                attach_stack_metadata_to_events(
                    next_step.id, current_flow.id, new_events
                )
                tracker.update_stack(step_stack)
                tracker.update_with_events(new_events)

    gathered_events = list(tracker.events)[number_of_initial_events:]
    if isinstance(step_result, PauseFlowReturnPrediction):
        prediction = step_result.action_prediction
        # make sure we really return all events that got created during the
        # step execution of all steps (not only the last one)
        prediction.events = gathered_events
        prediction.metadata = {
            ACTIVE_FLOW_METADATA_KEY: tracker.active_flow,
            STEP_ID_METADATA_KEY: tracker.current_step_id,
        }
        return prediction
    else:
        structlogger.warning("flow.step.execution.no_action")
        return FlowActionPrediction(None, 0.0, events=gathered_events)


def validate_collect_step(
    step: CollectInformationFlowStep,
    stack: DialogueStack,
    available_actions: List[str],
    slots: Dict[str, Slot],
    flow_name: str,
) -> bool:
    """Validate that a collect step can be executed.

    A collect step can be executed if either the `utter_ask` or the `action_ask` is
    defined in the domain. If neither is defined, the collect step can still be
    executed if the slot has an initial value defined in the domain, which would cause
    the step to be skipped.
    """
    slot = slots.get(step.collect)
    slot_has_initial_value_defined = slot and slot.initial_value is not None
    if (
        slot_has_initial_value_defined
        or step.utter in available_actions
        or step.collect_action in available_actions
    ):
        return True

    structlogger.error(
        "flow.step.run.collect_missing_utter_or_collect_action",
        slot_name=step.collect,
    )

    cancel_flow_and_push_internal_error(stack, flow_name)

    return False


def cancel_flow_and_push_internal_error(stack: DialogueStack, flow_name: str) -> None:
    """Cancel the top user flow and push the internal error pattern."""
    top_frame = stack.top()

    if isinstance(top_frame, BaseFlowStackFrame):
        # we need to first cancel the top user flow
        # because we cannot collect one of its slots
        # and therefore should not proceed with the flow
        # after triggering pattern_internal_error
        canceled_frames = CancelFlowCommand.select_canceled_frames(stack)
        stack.push(
            CancelPatternFlowStackFrame(
                canceled_name=flow_name,
                canceled_frames=canceled_frames,
            )
        )
    stack.push(InternalErrorPatternFlowStackFrame())


def attach_stack_metadata_to_events(
    step_id: str,
    flow_id: str,
    events: List[Event],
) -> None:
    """Attach the stack metadata to the events."""
    for event in events:
        event.metadata[STEP_ID_METADATA_KEY] = step_id
        event.metadata[ACTIVE_FLOW_METADATA_KEY] = flow_id


def run_step(
    step: FlowStep,
    flow: Flow,
    stack: DialogueStack,
    tracker: DialogueStateTracker,
    available_actions: List[str],
    flows: FlowsList,
    previous_step_id: str,
) -> FlowStepResult:
    """Run a single step of a flow.

    Returns the predicted action and a list of events that were generated
    during the step. The predicted action can be `None` if the step
    doesn't generate an action. The list of events can be empty if the
    step doesn't generate any events.

    Raises a `FlowException` if the step is invalid.

    Args:
        step: The step to run.
        flow: The flow that the step belongs to.
        stack: The stack that the flow is on.
        tracker: The tracker to run the step on.
        available_actions: The actions that are available in the domain.
        flows: All flows.
        previous_step_id: The ID of the previous step.

    Returns:
    A result of running the step describing where to transition to.
    """
    initial_events: List[Event] = []
    if previous_step_id == START_STEP:
        # if the previous step id is the start step, we need to add a flow
        # started event to the initial events.
        # we can't use the current step to check this, as the current step is the
        # first step in the flow -> other steps might link to this flow, so the
        # only reliable way to check if we are starting a new flow is checking for
        # the START_STEP meta step
        initial_events.append(FlowStarted(flow.id, metadata=stack.current_context()))

    # FLow does not start with collect step or we are not in collect information pattern
    if _first_step_is_not_collect(step, previous_step_id) and not (
        _in_collect_information_pattern(flow) or _in_pattern_user_silence(flow)
    ):
        _append_global_silence_timeout_event(initial_events, tracker)

    if isinstance(step, CollectInformationFlowStep):
        return _run_collect_information_step(
            available_actions,
            initial_events,
            stack,
            step,
            tracker,
            flow.readable_name(),
        )

    elif isinstance(step, ActionFlowStep):
        if not step.action:
            raise FlowException(f"Action not specified for step {step}")
        return _run_action_step(available_actions, initial_events, stack, step)

    elif isinstance(step, LinkFlowStep):
        return _run_link_step(initial_events, stack, step)

    elif isinstance(step, CallFlowStep):
        return _run_call_step(initial_events, stack, step)

    elif isinstance(step, SetSlotsFlowStep):
        return _run_set_slot_step(initial_events, step)

    elif isinstance(step, NoOperationFlowStep):
        structlogger.debug("flow.step.run.no_operation")
        return ContinueFlowWithNextStep(events=initial_events)

    elif isinstance(step, EndFlowStep):
        # If pattern collect information flow is ending,
        # we need to reset the silence timeout slot to its global value.
        if flow.id == FLOW_PATTERN_COLLECT_INFORMATION:
            _append_global_silence_timeout_event(initial_events, tracker)

        return _run_end_step(flow, flows, initial_events, stack, tracker)

    else:
        raise FlowException(f"Unknown flow step type {type(step)}")


def _first_step_is_not_collect(
    step: FlowStep,
    previous_step_id: str,
) -> bool:
    """Check if the first step is not a collect information step."""
    return (previous_step_id == START_STEP) and not isinstance(
        step, CollectInformationFlowStep
    )


def _in_collect_information_pattern(flow: Flow) -> bool:
    """Check if the current flow is a collect information pattern."""
    return flow.id == FLOW_PATTERN_COLLECT_INFORMATION


def _in_pattern_user_silence(flow: Flow) -> bool:
    """Check if the current flow is a user silence pattern."""
    return flow.id == FLOW_PATTERN_USER_SILENCE


def _run_end_step(
    flow: Flow,
    flows: FlowsList,
    initial_events: List[Event],
    stack: DialogueStack,
    tracker: DialogueStateTracker,
) -> FlowStepResult:
    # this is the end of the flow, so we'll pop it from the stack
    structlogger.debug("flow.step.run.flow_end")
    current_frame = stack.pop()
    trigger_pattern_completed(current_frame, stack, flows)
    resumed_events = trigger_pattern_continue_interrupted(
        current_frame, stack, flows, tracker
    )
    reset_events: List[Event] = reset_scoped_slots(current_frame, flow, tracker)
    return ContinueFlowWithNextStep(
        events=initial_events + reset_events + resumed_events, has_flow_ended=True
    )


def _run_set_slot_step(
    initial_events: List[Event], step: SetSlotsFlowStep
) -> FlowStepResult:
    structlogger.debug("flow.step.run.slot")
    slot_events: List[Event] = events_from_set_slots_step(step)
    return ContinueFlowWithNextStep(events=initial_events + slot_events)


def _run_call_step(
    initial_events: List[Event], stack: DialogueStack, step: CallFlowStep
) -> FlowStepResult:
    structlogger.debug("flow.step.run.call")
    stack.push(
        UserFlowStackFrame(
            flow_id=step.call,
            frame_type=FlowStackFrameType.CALL,
        ),
    )
    return ContinueFlowWithNextStep(events=initial_events)


def _run_link_step(
    initial_events: List[Event], stack: DialogueStack, step: LinkFlowStep
) -> FlowStepResult:
    structlogger.debug("flow.step.run.link")

    if step.link == RASA_PATTERN_HUMAN_HANDOFF:
        linked_stack_frame: DialogueStackFrame = HumanHandoffPatternFlowStackFrame()
    else:
        linked_stack_frame = UserFlowStackFrame(
            flow_id=step.link,
            frame_type=FlowStackFrameType.LINK,
        )

    stack.push(
        linked_stack_frame,
        # push this below the current stack frame so that we can
        # complete the current flow first and then continue with the
        # linked flow
        index=-1,
    )

    return ContinueFlowWithNextStep(events=initial_events)


def _run_action_step(
    available_actions: List[str],
    initial_events: List[Event],
    stack: DialogueStack,
    step: ActionFlowStep,
) -> FlowStepResult:
    context = {"context": stack.current_context()}
    action_name = render_template_variables(step.action, context)

    if action_name in available_actions:
        structlogger.debug("flow.step.run.action", context=context)
        return PauseFlowReturnPrediction(
            FlowActionPrediction(action_name, 1.0, events=initial_events)
        )
    else:
        if step.action != "validate_{{context.collect}}":
            # do not log about non-existing validation actions of collect steps
            utter_action_name = render_template_variables("{{context.utter}}", context)
            if utter_action_name not in available_actions:
                structlogger.warning("flow.step.run.action.unknown", action=action_name)
        return ContinueFlowWithNextStep(events=initial_events)


def _run_collect_information_step(
    available_actions: List[str],
    initial_events: List[Event],
    stack: DialogueStack,
    step: CollectInformationFlowStep,
    tracker: DialogueStateTracker,
    flow_name: str,
) -> FlowStepResult:
    is_step_valid = validate_collect_step(
        step, stack, available_actions, tracker.slots, flow_name
    )

    if not is_step_valid:
        # if we return any other FlowStepResult, the assistant will stay silent
        # instead of triggering the internal error pattern
        return ContinueFlowWithNextStep(events=initial_events)

    structlogger.debug("flow.step.run.collect")
    trigger_pattern_ask_collect_information(
        step.collect, stack, step.rejections, step.utter, step.collect_action
    )

    events: List[Event] = _events_for_collect_step_execution(step, tracker)
    return ContinueFlowWithNextStep(events=initial_events + events)


def _events_for_collect_step_execution(
    step: CollectInformationFlowStep, tracker: DialogueStateTracker
) -> List[Event]:
    """Create the events needed to prepare for the execution of a collect step."""
    # reset the slots that always need to be explicitly collected

    events = _silence_timeout_events_for_collect_step(step, tracker)

    slot = tracker.slots.get(step.collect, None)
    if slot and step.ask_before_filling:
        events.append(SlotSet(step.collect, None))

    return events


def _silence_timeout_events_for_collect_step(
    step: CollectInformationFlowStep, tracker: DialogueStateTracker
) -> List[Event]:
    events: List[Event] = []

    silence_timeout = (
        AvailableEndpoints.get_instance().interaction_handling.global_silence_timeout
    )

    if step.silence_timeout:
        structlogger.debug(
            "flow.step.run.adjusting_silence_timeout",
            duration=step.silence_timeout,
            collect=step.collect,
        )

        silence_timeout = step.silence_timeout
    else:
        structlogger.debug(
            "flow.step.run.reset_silence_timeout_to_global",
            duration=silence_timeout,
            collect=step.collect,
        )

    current_silence_timeout = tracker.get_slot(SILENCE_TIMEOUT_SLOT)

    if current_silence_timeout != silence_timeout:
        events.append(SlotSet(SILENCE_TIMEOUT_SLOT, silence_timeout))

    return events


def _append_global_silence_timeout_event(
    events: List[Event], tracker: DialogueStateTracker
) -> None:
    current_silence_timeout = tracker.get_slot(SILENCE_TIMEOUT_SLOT)
    global_silence_timeout = (
        AvailableEndpoints.get_instance().interaction_handling.global_silence_timeout
    )

    if current_silence_timeout != global_silence_timeout:
        events.append(
            SlotSet(
                SILENCE_TIMEOUT_SLOT,
                AvailableEndpoints.get_instance().interaction_handling.global_silence_timeout,
            )
        )



from rasa.dialogue_understanding.stack.dialogue_stack import DialogueStack
from rasa.shared.exceptions import RasaException


class FlowException(RasaException):
    """Exception that is raised when there is a problem with a flow."""

    pass


class FlowCircuitBreakerTrippedException(FlowException):
    """Exception that is raised when the flow circuit breaker tripped.

    The circuit breaker gets tripped when a flow seems to be stuck in
    executing steps and does not make any progress."""

    def __init__(
        self, dialogue_stack: DialogueStack, number_of_steps_taken: int
    ) -> None:
        """Creates a `FlowCircuitBreakerTrippedException`.

        Args:
            dialogue_stack: The dialogue stack.
            number_of_steps_taken: The number of steps that were taken.
        """
        super().__init__(
            f"Flow circuit breaker tripped after {number_of_steps_taken} steps. "
            "There appears to be an infinite loop in the flows."
        )
        self.dialogue_stack = dialogue_stack
        self.number_of_steps_taken = number_of_steps_taken


class NoNextStepInFlowException(FlowException):
    """Exception that is raised when there is no next step in a flow."""

    def __init__(self, dialogue_stack: DialogueStack) -> None:
        """Creates a `NoNextStepInFlowException`.

        Args:
            dialogue_stack: The dialogue stack.
        """
        super().__init__("No next step can be selected for the flow.")
        self.dialogue_stack = dialogue_stack



import argparse
from pathlib import Path

import rasa.shared.utils.cli
from rasa.shared.constants import (
    DEFAULT_CONFIG_PATH,
    DEFAULT_DATA_PATH,
    DEFAULT_ENDPOINTS_PATH,
)
from rasa.studio.constants import DOMAIN_FILENAME


def validate_argument_paths(args: argparse.Namespace) -> None:
    """Validates the paths provided in the command line arguments.

    Args:
        args: The command line arguments containing paths to validate.
    """

    def validate_path(arg_name: str, default: str) -> None:
        path_value = getattr(args, arg_name, None)
        if path_value and path_value != default:
            resolved_path = Path(path_value).resolve()
            if not resolved_path.exists():
                rasa.shared.utils.cli.print_error_and_exit(
                    f"{arg_name.capitalize()} file or directory "
                    f"'{path_value}' does not exist."
                )

    validate_path("domain", DOMAIN_FILENAME)
    validate_path("config", DEFAULT_CONFIG_PATH)
    validate_path("endpoints", DEFAULT_ENDPOINTS_PATH)
    validate_path("data", DEFAULT_DATA_PATH)



from typing import Any, Dict

import psutil

from rasa.core.nlg.contextual_response_rephraser import ContextualResponseRephraser
from rasa.core.policies.enterprise_search_policy import EnterpriseSearchPolicy
from rasa.core.policies.intentless_policy import IntentlessPolicy
from rasa.dialogue_understanding.generator import (
    CompactLLMCommandGenerator,
    LLMCommandGenerator,
    MultiStepLLMCommandGenerator,
    SearchReadyLLMCommandGenerator,
    SingleStepLLMCommandGenerator,
)
from rasa.tracing.constants import (
    COMPACT_LLM_COMMAND_GENERATOR_CPU_USAGE_METRIC_NAME,
    COMPACT_LLM_COMMAND_GENERATOR_LLM_RESPONSE_DURATION_METRIC_NAME,
    COMPACT_LLM_COMMAND_GENERATOR_MEMORY_USAGE_METRIC_NAME,
    COMPACT_LLM_COMMAND_GENERATOR_PROMPT_TOKEN_USAGE_METRIC_NAME,
    CONTEXTUAL_RESPONSE_REPHRASER_LLM_RESPONSE_DURATION_METRIC_NAME,
    ENTERPRISE_SEARCH_POLICY_CPU_USAGE_METRIC_NAME,
    ENTERPRISE_SEARCH_POLICY_LLM_RESPONSE_DURATION_METRIC_NAME,
    ENTERPRISE_SEARCH_POLICY_MEMORY_USAGE_METRIC_NAME,
    ENTERPRISE_SEARCH_POLICY_PROMPT_TOKEN_USAGE_METRIC_NAME,
    INTENTLESS_POLICY_LLM_RESPONSE_DURATION_METRIC_NAME,
    LLM_COMMAND_GENERATOR_CPU_USAGE_METRIC_NAME,
    LLM_COMMAND_GENERATOR_LLM_RESPONSE_DURATION_METRIC_NAME,
    LLM_COMMAND_GENERATOR_MEMORY_USAGE_METRIC_NAME,
    LLM_COMMAND_GENERATOR_PROMPT_TOKEN_USAGE_METRIC_NAME,
    MULTI_STEP_LLM_COMMAND_GENERATOR_CPU_USAGE_METRIC_NAME,
    MULTI_STEP_LLM_COMMAND_GENERATOR_LLM_RESPONSE_DURATION_METRIC_NAME,
    MULTI_STEP_LLM_COMMAND_GENERATOR_MEMORY_USAGE_METRIC_NAME,
    MULTI_STEP_LLM_COMMAND_GENERATOR_PROMPT_TOKEN_USAGE_METRIC_NAME,
    PROMPT_TOKEN_LENGTH_ATTRIBUTE_NAME,
    RASA_CLIENT_REQUEST_BODY_SIZE_METRIC_NAME,
    RASA_CLIENT_REQUEST_DURATION_METRIC_NAME,
    REQUEST_BODY_SIZE_IN_BYTES_ATTRIBUTE_NAME,
    SEARCH_READY_LLM_COMMAND_GENERATOR_CPU_USAGE_METRIC_NAME,
    SEARCH_READY_LLM_COMMAND_GENERATOR_LLM_RESPONSE_DURATION_METRIC_NAME,
    SEARCH_READY_LLM_COMMAND_GENERATOR_MEMORY_USAGE_METRIC_NAME,
    SEARCH_READY_LLM_COMMAND_GENERATOR_PROMPT_TOKEN_USAGE_METRIC_NAME,
    SINGLE_STEP_LLM_COMMAND_GENERATOR_CPU_USAGE_METRIC_NAME,
    SINGLE_STEP_LLM_COMMAND_GENERATOR_LLM_RESPONSE_DURATION_METRIC_NAME,
    SINGLE_STEP_LLM_COMMAND_GENERATOR_MEMORY_USAGE_METRIC_NAME,
    SINGLE_STEP_LLM_COMMAND_GENERATOR_PROMPT_TOKEN_USAGE_METRIC_NAME,
)
from rasa.tracing.metric_instrument_provider import MetricInstrumentProvider
from rasa.utils.endpoints import EndpointConfig


def record_llm_based_component_cpu_usage(
    metric_instrument_provider: MetricInstrumentProvider,
    metric_name: str,
) -> None:
    """Record CPU usage as a percentage.

    The recording is done by the opentelemetry.metrics.Histogram instrument.
    This instrument is registered to the MetricInstrumentProvider internal singleton.

    :param metric_instrument_provider: The MetricInstrumentProvider instance
    :param metric_name: The name of the metric instrument
    :return: None
    """
    metric_instrument = metric_instrument_provider.get_instrument(metric_name)
    if not metric_instrument:
        return None

    cpu_usage = psutil.cpu_percent()
    metric_instrument.record(amount=cpu_usage)

    return None


def record_llm_based_component_memory_usage(
    metric_instrument_provider: MetricInstrumentProvider,
    metric_name: str,
) -> None:
    """Record memory usage as a percentage.

    The recording is done by the opentelemetry.metrics.Histogram instrument.
    This instrument is registered to the MetricInstrumentProvider internal singleton.

    :param metric_instrument_provider: The MetricInstrumentProvider instance
    :param metric_name: The name of the metric instrument
    :return: None
    """
    metric_instrument = metric_instrument_provider.get_instrument(metric_name)
    if not metric_instrument:
        return None

    memory_usage = psutil.virtual_memory().percent
    metric_instrument.record(amount=memory_usage)

    return None


def record_llm_based_component_prompt_token(
    metric_instrument_provider: MetricInstrumentProvider,
    attributes: Dict[str, Any],
    metric_name: str,
) -> None:
    """
    Record prompt token length.

    The recording is done by the opentelemetry.metrics.Histogram instrument.
    This instrument is registered to the MetricInstrumentProvider internal singleton.

    :param metric_instrument_provider: The MetricInstrumentProvider instance
    :param attributes: Extracted tracing attributes
    :param metric_name: The name of the metric instrument
    :return: None
    """
    if PROMPT_TOKEN_LENGTH_ATTRIBUTE_NAME not in attributes:
        return None

    metric_instrument = metric_instrument_provider.get_instrument(metric_name)
    if not metric_instrument:
        return None

    prompt_tokens_len = attributes[PROMPT_TOKEN_LENGTH_ATTRIBUTE_NAME]

    try:
        prompt_tokens_len = int(prompt_tokens_len)
    except ValueError:
        return None

    metric_instrument.record(
        amount=prompt_tokens_len,
    )

    return None


def record_llm_command_generator_metrics(attributes: Dict[str, Any]) -> None:
    """
    Record measurements for LLMCommandGenerator specific metrics.

    The recording is done by the opentelemetry.metrics.Histogram instruments.
    These instruments are registered to the MetricInstrumentProvider internal singleton.

    :param attributes: Extracted tracing attributes
    :return: None
    """
    instrument_provider = MetricInstrumentProvider()

    if not instrument_provider.instruments:
        return None

    record_llm_based_component_cpu_usage(
        instrument_provider, LLM_COMMAND_GENERATOR_CPU_USAGE_METRIC_NAME
    )
    record_llm_based_component_memory_usage(
        instrument_provider, LLM_COMMAND_GENERATOR_MEMORY_USAGE_METRIC_NAME
    )
    record_llm_based_component_prompt_token(
        instrument_provider,
        attributes,
        LLM_COMMAND_GENERATOR_PROMPT_TOKEN_USAGE_METRIC_NAME,
    )
    return None


def record_single_step_llm_command_generator_metrics(
    attributes: Dict[str, Any],
) -> None:
    """
    Record measurements for SingleStepLLMCommandGenerator specific metrics.

    The recording is done by the opentelemetry.metrics.Histogram instruments.
    These instruments are registered to the MetricInstrumentProvider internal singleton.

    :param attributes: Extracted tracing attributes
    :return: None
    """
    instrument_provider = MetricInstrumentProvider()

    if not instrument_provider.instruments:
        return None

    record_llm_based_component_cpu_usage(
        instrument_provider, SINGLE_STEP_LLM_COMMAND_GENERATOR_CPU_USAGE_METRIC_NAME
    )
    record_llm_based_component_memory_usage(
        instrument_provider, SINGLE_STEP_LLM_COMMAND_GENERATOR_MEMORY_USAGE_METRIC_NAME
    )
    record_llm_based_component_prompt_token(
        instrument_provider,
        attributes,
        SINGLE_STEP_LLM_COMMAND_GENERATOR_PROMPT_TOKEN_USAGE_METRIC_NAME,
    )
    return None


def record_compact_llm_command_generator_metrics(
    attributes: Dict[str, Any],
) -> None:
    """
    Record measurements for CompactLLMCommandGenerator specific metrics.

    The recording is done by the opentelemetry.metrics.Histogram instruments.
    These instruments are registered to the MetricInstrumentProvider internal singleton.

    :param attributes: Extracted tracing attributes
    :return: None
    """
    instrument_provider = MetricInstrumentProvider()

    if not instrument_provider.instruments:
        return None

    record_llm_based_component_cpu_usage(
        instrument_provider, COMPACT_LLM_COMMAND_GENERATOR_CPU_USAGE_METRIC_NAME
    )
    record_llm_based_component_memory_usage(
        instrument_provider, COMPACT_LLM_COMMAND_GENERATOR_MEMORY_USAGE_METRIC_NAME
    )
    record_llm_based_component_prompt_token(
        instrument_provider,
        attributes,
        COMPACT_LLM_COMMAND_GENERATOR_PROMPT_TOKEN_USAGE_METRIC_NAME,
    )
    return None


def record_search_ready_llm_command_generator_metrics(
    attributes: Dict[str, Any],
) -> None:
    """
    Record measurements for SearchReadyLLMCommandGenerator specific metrics.

    The recording is done by the opentelemetry.metrics.Histogram instruments.
    These instruments are registered to the MetricInstrumentProvider internal singleton.

    :param attributes: Extracted tracing attributes
    :return: None
    """
    instrument_provider = MetricInstrumentProvider()

    if not instrument_provider.instruments:
        return None

    record_llm_based_component_cpu_usage(
        instrument_provider, SEARCH_READY_LLM_COMMAND_GENERATOR_CPU_USAGE_METRIC_NAME
    )
    record_llm_based_component_memory_usage(
        instrument_provider, SEARCH_READY_LLM_COMMAND_GENERATOR_MEMORY_USAGE_METRIC_NAME
    )
    record_llm_based_component_prompt_token(
        instrument_provider,
        attributes,
        SEARCH_READY_LLM_COMMAND_GENERATOR_PROMPT_TOKEN_USAGE_METRIC_NAME,
    )
    return None


def record_multi_step_llm_command_generator_metrics(attributes: Dict[str, Any]) -> None:
    """
    Record measurements for MultiStepLLMCommandGenerator specific metrics.

    The recording is done by the opentelemetry.metrics.Histogram instruments.
    These instruments are registered to the MetricInstrumentProvider internal singleton.

    :param attributes: Extracted tracing attributes
    :return: None
    """
    instrument_provider = MetricInstrumentProvider()

    if not instrument_provider.instruments:
        return None

    record_llm_based_component_cpu_usage(
        instrument_provider, MULTI_STEP_LLM_COMMAND_GENERATOR_CPU_USAGE_METRIC_NAME
    )
    record_llm_based_component_memory_usage(
        instrument_provider, MULTI_STEP_LLM_COMMAND_GENERATOR_MEMORY_USAGE_METRIC_NAME
    )
    record_llm_based_component_prompt_token(
        instrument_provider,
        attributes,
        MULTI_STEP_LLM_COMMAND_GENERATOR_PROMPT_TOKEN_USAGE_METRIC_NAME,
    )
    return None


def record_enterprise_search_policy_metrics(attributes: Dict[str, Any]) -> None:
    """
    Record measurements for EnterpriseSearchPolicy specific metrics.

    The recording is done by the opentelemetry.metrics.Histogram instruments.
    These instruments are registered to the MetricInstrumentProvider internal singleton.

    :param attributes: Extracted tracing attributes
    :return: None
    """
    instrument_provider = MetricInstrumentProvider()

    if not instrument_provider.instruments:
        return None

    record_llm_based_component_cpu_usage(
        instrument_provider, ENTERPRISE_SEARCH_POLICY_CPU_USAGE_METRIC_NAME
    )
    record_llm_based_component_memory_usage(
        instrument_provider, ENTERPRISE_SEARCH_POLICY_MEMORY_USAGE_METRIC_NAME
    )
    record_llm_based_component_prompt_token(
        instrument_provider,
        attributes,
        ENTERPRISE_SEARCH_POLICY_PROMPT_TOKEN_USAGE_METRIC_NAME,
    )
    return None


def record_callable_duration_metrics(
    self: Any, start_time: int, end_time: int, **kwargs: Any
) -> None:
    """
    Record duration of instrumented method calls invoked for the following components:
    - LLMCommandGenerator
    - SingleStepLLMCommandGenerator
    - CompactLLMCommandGenerator
    - MultiStepLLMCommandGenerator
    - EnterpriseSearchPolicy
    - IntentlessPolicy
    - ContextualResponseRephraser
    - EndpointConfig

    :param self: The instance on which the instrumented method is called on.
    :param start_time: Start time measured by time.perf_counter_ns()
    before the method is called.
    :param end_time: End time measured by time.perf_counter_ns()
    after the method is called.
    :return: None
    """
    instrument_provider = MetricInstrumentProvider()
    if not instrument_provider.instruments:
        return None

    metric_instrument = None
    attributes = {}

    if type(self) == LLMCommandGenerator:
        metric_instrument = instrument_provider.get_instrument(
            LLM_COMMAND_GENERATOR_LLM_RESPONSE_DURATION_METRIC_NAME
        )

    if type(self) == SingleStepLLMCommandGenerator:
        metric_instrument = instrument_provider.get_instrument(
            SINGLE_STEP_LLM_COMMAND_GENERATOR_LLM_RESPONSE_DURATION_METRIC_NAME
        )

    if type(self) == CompactLLMCommandGenerator:
        metric_instrument = instrument_provider.get_instrument(
            COMPACT_LLM_COMMAND_GENERATOR_LLM_RESPONSE_DURATION_METRIC_NAME
        )

    if type(self) == SearchReadyLLMCommandGenerator:
        metric_instrument = instrument_provider.get_instrument(
            SEARCH_READY_LLM_COMMAND_GENERATOR_LLM_RESPONSE_DURATION_METRIC_NAME
        )

    if type(self) == MultiStepLLMCommandGenerator:
        metric_instrument = instrument_provider.get_instrument(
            MULTI_STEP_LLM_COMMAND_GENERATOR_LLM_RESPONSE_DURATION_METRIC_NAME
        )

    if isinstance(self, EnterpriseSearchPolicy):
        metric_instrument = instrument_provider.get_instrument(
            ENTERPRISE_SEARCH_POLICY_LLM_RESPONSE_DURATION_METRIC_NAME
        )

    if isinstance(self, IntentlessPolicy):
        metric_instrument = instrument_provider.get_instrument(
            INTENTLESS_POLICY_LLM_RESPONSE_DURATION_METRIC_NAME
        )

    if isinstance(self, ContextualResponseRephraser):
        metric_instrument = instrument_provider.get_instrument(
            CONTEXTUAL_RESPONSE_REPHRASER_LLM_RESPONSE_DURATION_METRIC_NAME
        )

    if isinstance(self, EndpointConfig):
        metric_instrument = instrument_provider.get_instrument(
            RASA_CLIENT_REQUEST_DURATION_METRIC_NAME
        )
        attributes = {"url": kwargs.get("url")}

    if not metric_instrument:
        return None

    timedelta_ms = (end_time - start_time) / 1000000
    metric_instrument.record(amount=timedelta_ms, attributes=attributes)


def record_request_size_in_bytes(attributes: Dict[str, Any]) -> None:
    """
    Record endpoint request size in bytes for EndpointConfig class.

    The recording is done by the opentelemetry.metrics.Histogram instruments.
    These instruments are registered to the MetricInstrumentProvider internal singleton.

    :param attributes: Extracted tracing attributes
    :return: None
    """

    metric_instrument_provider = MetricInstrumentProvider()

    if not metric_instrument_provider.instruments:
        return None

    metric_instrument = metric_instrument_provider.get_instrument(
        RASA_CLIENT_REQUEST_BODY_SIZE_METRIC_NAME
    )
    if not metric_instrument:
        return None

    request_body_size = attributes.pop(REQUEST_BODY_SIZE_IN_BYTES_ATTRIBUTE_NAME, 0)
    metric_instrument.record(amount=request_body_size, attributes=attributes)



from rasa.dialogue_understanding.generator.command_generator import CommandGenerator
from rasa.dialogue_understanding.generator.llm_based_command_generator import (
    LLMBasedCommandGenerator,
)
from rasa.dialogue_understanding.generator.llm_command_generator import (
    LLMCommandGenerator,
)
from rasa.dialogue_understanding.generator.multi_step.multi_step_llm_command_generator import (  # noqa: E501
    MultiStepLLMCommandGenerator,
)
from rasa.dialogue_understanding.generator.single_step.compact_llm_command_generator import (  # noqa: E501
    CompactLLMCommandGenerator,
)
from rasa.dialogue_understanding.generator.single_step.search_ready_llm_command_generator import (  # noqa: E501
    SearchReadyLLMCommandGenerator,
)
from rasa.dialogue_understanding.generator.single_step.single_step_llm_command_generator import (  # noqa: E501
    SingleStepLLMCommandGenerator,
)

__all__ = [
    "CommandGenerator",
    "LLMBasedCommandGenerator",
    "LLMCommandGenerator",
    "MultiStepLLMCommandGenerator",
    "SingleStepLLMCommandGenerator",
    "CompactLLMCommandGenerator",
    "SearchReadyLLMCommandGenerator",
]



import os
import shlex
import subprocess
import uuid
from dataclasses import dataclass
from typing import List

import structlog

from rasa.__main__ import main
from rasa.dialogue_understanding.constants import (
    RASA_RECORD_COMMANDS_AND_PROMPTS_ENV_VAR_NAME,
)
from rasa.model_manager import config
from rasa.model_manager.utils import ensure_base_directory_exists, logs_path

structlogger = structlog.get_logger(__name__)

warm_rasa_processes: List["WarmRasaProcess"] = []

NUMBER_OF_INITIAL_PROCESSES = 3


@dataclass
class WarmRasaProcess:
    """Data class to store a warm Rasa process.

    A "warm" Rasa process is one where we've done the heavy lifting of
    importing key modules ahead of time (e.g. litellm). This is to avoid
    long import times when we actually want to run a command.

    This is a started process waiting for a Rasa CLI command. It's
    output is stored in a log file identified by `log_id`.
    """

    process: subprocess.Popen
    log_id: str


def _create_warm_rasa_process() -> WarmRasaProcess:
    """Create a new warm Rasa process."""
    command = [
        config.RASA_PYTHON_PATH,
        "-m",
        "rasa.model_manager.warm_rasa_process",
    ]

    envs = os.environ.copy()
    envs[RASA_RECORD_COMMANDS_AND_PROMPTS_ENV_VAR_NAME] = "true"

    log_id = uuid.uuid4().hex
    log_path = logs_path(log_id)

    ensure_base_directory_exists(log_path)

    process = subprocess.Popen(
        command,
        stdout=open(log_path, "w"),
        stderr=subprocess.STDOUT,
        stdin=subprocess.PIPE,
        env=envs,
    )

    structlogger.debug(
        "model_trainer.created_warm_rasa_process",
        pid=process.pid,
        command=command,
        log_path=log_path,
    )

    return WarmRasaProcess(process=process, log_id=log_id)


def initialize_warm_rasa_process() -> None:
    """Initialize the warm Rasa processes."""
    global warm_rasa_processes
    for _ in range(NUMBER_OF_INITIAL_PROCESSES):
        warm_rasa_processes.append(_create_warm_rasa_process())


def shutdown_warm_rasa_processes() -> None:
    """Shutdown all warm Rasa processes."""
    global warm_rasa_processes
    for warm_rasa_process in warm_rasa_processes:
        warm_rasa_process.process.terminate()
    warm_rasa_processes = []


def start_rasa_process(cwd: str, arguments: List[str]) -> WarmRasaProcess:
    """Start a Rasa process.

    This will start a Rasa process with the given current working directory
    and arguments. The process will be a warm one, meaning that it has already
    imported all necessary modules.
    """
    warm_rasa_process = _get_warm_rasa_process()
    _pass_arguments_to_process(warm_rasa_process.process, cwd, arguments)
    return warm_rasa_process


def _get_warm_rasa_process() -> WarmRasaProcess:
    """Get a warm Rasa process.

    This will return a warm Rasa process from the pool and create a
    new one to replace it.
    """
    global warm_rasa_processes

    if not warm_rasa_processes:
        warm_rasa_processes = [_create_warm_rasa_process()]

    previous_warm_rasa_process = warm_rasa_processes.pop(0)

    if previous_warm_rasa_process.process.poll() is not None:
        # process has finished (for some reason...)
        # back up plan is to create a new one on the spot.
        # this should not happen, but let's be safe
        structlogger.warning(
            "model_trainer.warm_rasa_process_finished_unexpectedly",
            pid=previous_warm_rasa_process.process.pid,
        )
        previous_warm_rasa_process = _create_warm_rasa_process()

    warm_rasa_processes.append(_create_warm_rasa_process())
    return previous_warm_rasa_process


def _pass_arguments_to_process(
    process: subprocess.Popen, cwd: str, arguments: List[str]
) -> None:
    """Pass arguments to a warm Rasa process.

    The process is waiting for input on stdin. We pass the current working
    directory and the arguments to run a Rasa CLI command.
    """
    arguments_string = " ".join(arguments)
    # send arguments to stdin
    process.stdin.write(cwd.encode())  # type: ignore[union-attr]
    process.stdin.write("\n".encode())  # type: ignore[union-attr]
    process.stdin.write(arguments_string.encode())  # type: ignore[union-attr]
    process.stdin.write("\n".encode())  # type: ignore[union-attr]
    process.stdin.flush()  # type: ignore[union-attr]


def warmup() -> None:
    """Import all necessary modules to warm up the process.

    This should include all the modules that take a long time to import.
    We import them now, so that the training / deployment can later
    directly start.
    """
    try:
        import langchain  # noqa: F401
        import litellm  # noqa: F401
        import matplotlib  # noqa: F401
        import numpy  # noqa: F401
        import pandas  # noqa: F401
        import presidio_analyzer  # noqa: F401
        import spacy  # noqa: F401
        import tensorflow  # noqa: F401

        import rasa.validator  # noqa: F401
    except ImportError:
        pass


def warm_rasa_main() -> None:
    """Entry point for processes waiting for their command to run.

    The process will wait for the current working directory and the command
    to run. These will be send on stdin by the parent process. After receiving
    the input, we will kick things of starting or running a bot.

    Uses the normal Rasa CLI entry point (e.g. `rasa train --data ...`).
    """
    warmup()

    cwd = input()

    # this should be `train --data ...` or similar
    cli_arguments_str = input()
    # splits the arguments string into a list of arguments as expected by `argparse`
    arguments = shlex.split(cli_arguments_str)

    # needed to make sure the passed arguments are relative to the working directory
    os.chdir(cwd)

    main(arguments)


if __name__ == "__main__":
    warm_rasa_main()



from __future__ import annotations

import os
import re
from typing import Any, Dict, Optional

import structlog

from rasa.shared.constants import (
    API_KEY,
    AZURE_API_BASE_ENV_VAR,
    AZURE_API_KEY_ENV_VAR,
    AZURE_API_TYPE_ENV_VAR,
    AZURE_API_VERSION_ENV_VAR,
    AZURE_OPENAI_PROVIDER,
    OPENAI_API_BASE_ENV_VAR,
    OPENAI_API_KEY_ENV_VAR,
    OPENAI_API_TYPE_ENV_VAR,
    OPENAI_API_VERSION_ENV_VAR,
)
from rasa.shared.exceptions import ProviderClientValidationError
from rasa.shared.providers._configs.azure_openai_client_config import (
    AzureEntraIDOAuthConfig,
    AzureOpenAIClientConfig,
)
from rasa.shared.providers._utils import validate_azure_client_setup
from rasa.shared.providers.constants import (
    DEFAULT_AZURE_API_KEY_NAME,
    LITE_LLM_API_BASE_FIELD,
    LITE_LLM_API_KEY_FIELD,
    LITE_LLM_API_VERSION_FIELD,
    LITE_LLM_AZURE_AD_TOKEN,
)
from rasa.shared.providers.llm._base_litellm_client import _BaseLiteLLMClient
from rasa.shared.utils.io import raise_deprecation_warning

structlogger = structlog.get_logger()

AZURE_CLIENT_ID = "AZURE_CLIENT_ID"
AZURE_CLIENT_SECRET = "AZURE_CLIENT_SECRET"
AZURE_TENANT_ID = "AZURE_TENANT_ID"
CLIENT_SECRET_VARS = (AZURE_CLIENT_ID, AZURE_CLIENT_SECRET, AZURE_TENANT_ID)

AZURE_CLIENT_CERTIFICATE_PATH = "AZURE_CLIENT_CERTIFICATE_PATH"
AZURE_CLIENT_CERTIFICATE_PASSWORD = "AZURE_CLIENT_CERTIFICATE_PASSWORD"
AZURE_CLIENT_SEND_CERTIFICATE_CHAIN = "AZURE_CLIENT_SEND_CERTIFICATE_CHAIN"
CERT_VARS = (AZURE_CLIENT_ID, AZURE_CLIENT_CERTIFICATE_PATH, AZURE_TENANT_ID)


class AzureADConfig:
    def __init__(
        self, client_id: str, client_secret: str, tenant_id: str, scopes: str
    ) -> None:
        self.scopes = scopes


class AzureOpenAILLMClient(_BaseLiteLLMClient):
    """A client for interfacing with Azure's OpenAI LLM deployments.

    Parameters:
        deployment (str): The deployment name.
        model (Optional[str]): The name of the deployed model.
        api_type: (Optional[str]): The api type. If not provided, it will be set via
            environment variable.
        api_base (Optional[str]): The base URL for the API endpoints. If not provided,
            it will be set via environment variables.
        api_version (Optional[str]): The version of the API to use. If not provided,
            it will be set via environment variable.

        kwargs (Optional[Dict[str, Any]]): Optional configuration parameters specific
            to the model deployment.

    Raises:
        ProviderClientValidationError: If validation of the client setup fails.
        DeprecationWarning: If deprecated environment variables are used for
            configuration.
    """

    def __init__(
        self,
        deployment: str,
        model: Optional[str] = None,
        api_type: Optional[str] = None,
        api_base: Optional[str] = None,
        api_version: Optional[str] = None,
        oauth: Optional[AzureEntraIDOAuthConfig] = None,
        **kwargs: Any,
    ):
        super().__init__()  # type: ignore
        self._deployment = deployment
        self._model = model
        self._extra_parameters = kwargs or {}

        # Set api_base with the following priority:
        # parameter -> Azure Env Var -> (deprecated) OpenAI Env Var
        self._api_base = (
            api_base
            or os.getenv(AZURE_API_BASE_ENV_VAR)
            or os.getenv(OPENAI_API_BASE_ENV_VAR)
        )

        # Set api_version with the following priority:
        # parameter -> Azure Env Var -> (deprecated) OpenAI Env Var
        self._api_version = (
            api_version
            or os.getenv(AZURE_API_VERSION_ENV_VAR)
            or os.getenv(OPENAI_API_VERSION_ENV_VAR)
        )

        # Not used by LiteLLM, here for backward compatibility
        self._api_type = (
            api_type
            or os.getenv(AZURE_API_TYPE_ENV_VAR)
            or os.getenv(OPENAI_API_TYPE_ENV_VAR)
        )

        os.unsetenv("OPENAI_API_KEY")
        os.unsetenv("AZURE_API_KEY")

        self._oauth = oauth

        if self._oauth:
            os.unsetenv(DEFAULT_AZURE_API_KEY_NAME)
            os.unsetenv(AZURE_API_KEY_ENV_VAR)
            os.unsetenv(OPENAI_API_KEY_ENV_VAR)
        self._api_key_env_var = (
            self._resolve_api_key_env_var() if not self._oauth else None
        )

        # Run helper function to check and raise deprecation warning if
        # deprecated environment variables were used for initialization of the
        # client settings
        self._raise_evn_var_deprecation_warnings()

        # validate the client settings
        self.validate_client_setup()

    def _raise_evn_var_deprecation_warnings(self) -> None:
        """Helper function to check and raise deprecation warning if
        deprecated environment variables were used for initialization of
        some client settings.
        """
        deprecation_mapping = {
            "API Base": {
                "current_value": self.api_base,
                "env_var": AZURE_API_BASE_ENV_VAR,
                "deprecated_var": OPENAI_API_BASE_ENV_VAR,
            },
            "API Version": {
                "current_value": self.api_version,
                "env_var": AZURE_API_VERSION_ENV_VAR,
                "deprecated_var": OPENAI_API_VERSION_ENV_VAR,
            },
        }

        deprecation_warning_message = (
            "Usage of {deprecated_env_var} environment "
            "variable for setting the {setting} for Azure "
            "OpenAI client is deprecated and will be removed "
            "in 4.0.0. "
        )
        deprecation_warning_replacement_message = (
            "Please use {env_var} environment variable."
        )

        for setting in deprecation_mapping.keys():
            current_value = deprecation_mapping[setting]["current_value"]
            env_var = deprecation_mapping[setting]["env_var"]
            deprecated_var = deprecation_mapping[setting]["deprecated_var"]

            # Value is set through the non-deprecated env var
            if current_value == os.getenv(env_var):
                continue

            # Value is set through the deprecated env var
            if current_value == os.getenv(deprecated_var):
                message = deprecation_warning_message.format(
                    setting=setting, deprecated_env_var=deprecated_var
                )
                if env_var is not None:
                    message += deprecation_warning_replacement_message.format(
                        env_var=env_var
                    )
                raise_deprecation_warning(message=message)

    def _resolve_api_key_env_var(self) -> str:
        """Resolves the environment variable to use for the API key.

        Returns:
            str: The env variable in dollar syntax format to use for the API key.
        """
        if API_KEY in self._extra_parameters:
            # API key is set to an env var in the config itself
            # in case the model is defined in the endpoints.yml
            return self._extra_parameters[API_KEY]

        if os.getenv(AZURE_API_KEY_ENV_VAR) is not None:
            return f"${{{DEFAULT_AZURE_API_KEY_NAME}}}"

        if os.getenv(OPENAI_API_KEY_ENV_VAR) is not None:
            # API key can be set through OPENAI_API_KEY too,
            # because of the backward compatibility
            raise_deprecation_warning(
                message=(
                    f"Usage of '{OPENAI_API_KEY_ENV_VAR}' environment variable "
                    "for setting the API key for Azure OpenAI "
                    "client is deprecated and will be removed "
                    f"in 4.0.0. Please use '{AZURE_API_KEY_ENV_VAR}' "
                    "environment variable."
                )
            )
            return "${OPENAI_API_KEY}"

        structlogger.error(
            "azure_openai_llm_client.api_key_not_set",
            event_info=(
                "API key not set, it is required for API calls. "
                f"Set it either via the environment variable"
                f"'{AZURE_API_KEY_ENV_VAR}' or directly"
                f"via the config key '{API_KEY}'."
            ),
        )
        raise ProviderClientValidationError(
            f"Missing required environment variable/config key '{API_KEY}' for "
            f"API calls."
        )

    @classmethod
    def from_config(cls, config: Dict[str, Any]) -> AzureOpenAILLMClient:
        """Initializes the client from given configuration.

        Args:
            config (Dict[str, Any]): Configuration.

        Raises:
            ValueError:
                If any of the required configuration keys are missing.
                If `api_type` has a value different from `azure`.

        Returns:
            AzureOpenAILLMClient: Initialized client.
        """
        try:
            azure_openai_config = AzureOpenAIClientConfig.from_dict(config)
        except ValueError as e:
            message = "Cannot instantiate a client from the passed configuration."
            structlogger.error(
                "azure_openai_llm_client.from_config.error",
                message=message,
                config=config,
                original_error=e,
            )
            raise

        return cls(
            deployment=azure_openai_config.deployment,
            model=azure_openai_config.model,
            api_type=azure_openai_config.api_type,
            api_base=azure_openai_config.api_base,
            api_version=azure_openai_config.api_version,
            oauth=azure_openai_config.oauth,
            **azure_openai_config.extra_parameters,
        )

    @property
    def config(self) -> dict:
        """Returns the configuration for that the llm client
        in dictionary form.
        """
        config = AzureOpenAIClientConfig(
            deployment=self._deployment,
            model=self._model,
            api_base=self._api_base,
            api_version=self._api_version,
            api_type=self._api_type,
            oauth=self._oauth,
            extra_parameters=self._extra_parameters,
        )
        return config.to_dict()

    @property
    def deployment(self) -> str:
        return self._deployment

    @property
    def model(self) -> Optional[str]:
        """Returns the name of the model deployed on Azure."""
        return self._model

    @property
    def api_base(self) -> Optional[str]:
        """Returns the API base URL for the Azure OpenAI llm client."""
        return self._api_base

    @property
    def api_version(self) -> Optional[str]:
        """Returns the API version for the Azure OpenAI llm client."""
        return self._api_version

    @property
    def api_type(self) -> Optional[str]:
        return self._api_type

    @property
    def _litellm_model_name(self) -> str:
        """Returns the value of LiteLLM's model parameter to be used in
        completion/acompletion in LiteLLM format:

        <provider>/<model or deployment name>
        """
        regex_pattern = rf"^{AZURE_OPENAI_PROVIDER}/"
        if not re.match(regex_pattern, self._deployment):
            return f"{AZURE_OPENAI_PROVIDER}/{self._deployment}"
        return self._deployment

    @property
    def _litellm_extra_parameters(self) -> Dict[str, Any]:
        return self._extra_parameters

    @property
    def _completion_fn_args(self) -> Dict[str, Any]:
        """Returns the completion arguments for invoking a call through
        LiteLLM's completion functions.
        """
        # Set the API key env var to None if OAuth is used
        auth_parameter: Dict[str, str] = {}

        if self._oauth:
            auth_parameter = {
                **auth_parameter,
                LITE_LLM_AZURE_AD_TOKEN: self._oauth.get_bearer_token(),
            }
        elif self._api_key_env_var:
            auth_parameter = {LITE_LLM_API_KEY_FIELD: self._api_key_env_var}

        fn_args = super()._completion_fn_args
        fn_args.update(
            {
                LITE_LLM_API_BASE_FIELD: self.api_base,
                LITE_LLM_API_VERSION_FIELD: self.api_version,
                **auth_parameter,
            }
        )
        return fn_args

    def validate_client_setup(self) -> None:
        """Validates that all required configuration parameters are set."""

        return validate_azure_client_setup(
            api_base=self.api_base,
            api_version=self.api_version,
            deployment=self.deployment,
        )
